"""
Type stubs for UnitHandle - Provides strict IDE validation and autocomplete.
Unlike the runtime class, this stub does not permit arbitrary attribute assignment.
"""
from typing import List, Optional, Tuple, Any, Union

from aoe2_genie_tooling.Base.workspace import GenieWorkspace
from aoe2_genie_tooling.Units.task_builder import TaskBuilder

from aoe2_genie_tooling.Units.handles import (
    TaskHandle, AttackHandle, ArmourHandle, DamageGraphicHandle,
    TrainLocationHandle, DropSiteHandle
)

from aoe2_genie_tooling.Units.wrappers import (
    CombatWrapper,
    CreationWrapper,
    MovementWrapper,
    BehaviorWrapper,
    ProjectileWrapper,
    BuildingWrapper
)

from aoe2_genie_tooling.Units.unit_collections import (
    TasksManager,
    AttacksManager,
    ArmoursManager,
    DamageGraphicsManager,
    TrainLocationsManager,
    DropSitesManager,
    AnnexesManager,
    CostsManager,
    ResourcesManager
)

class UnitHandle:
    __slots__ = (
        "_workspace", "_unit_id", "_civ_ids", "_units_cache",
        "_combat_cache", "_creation_cache", "_cost_cache", "_movement_cache",
        "_behavior_cache", "_projectile_cache", "_building_cache",
        "_tasks_cache", "_attacks_cache", "_armours_cache", 
        "_damage_graphics_cache", "_train_locations_cache", 
        "_annexes_cache", "_drop_sites_cache", "_costs_cache", "_resources_cache",
    )

    # --------------------------------------------------------------------------
    # Core Attributes
    # --------------------------------------------------------------------------
    @property
    def id(self) -> int: ...
    @property
    def unit_id(self) -> int: ...

    @property
    def name(self) -> str: ...
    @name.setter
    def name(self, value: str) -> None: ...

    @property
    def type_(self) -> int: ...
    @type_.setter
    def type_(self, value: int) -> None: ...

    def change_unit_type(self, new_type: int) -> None:
        """Safely change unit type (structures sync at save time)."""
        ...

    @property
    def enabled(self) -> int: ...
    @enabled.setter
    def enabled(self, value: int) -> None: ...

    @property
    def disabled(self) -> int: ...
    @disabled.setter
    def disabled(self, value: int) -> None: ...

    @property
    def class_(self) -> int: ...
    @class_.setter
    def class_(self, value: int) -> None: ...

    @property
    def hit_points(self) -> int: ...
    @hit_points.setter
    def hit_points(self, value: int) -> None: ...

    @property
    def line_of_sight(self) -> float: ...
    @line_of_sight.setter
    def line_of_sight(self, value: float) -> None: ...

    @property
    def garrison_capacity(self) -> int: ...
    @garrison_capacity.setter
    def garrison_capacity(self, value: int) -> None: ...

    @property
    def speed(self) -> float: ...
    @speed.setter
    def speed(self, value: float) -> None: ...

    @property
    def name_str_id(self) -> int: ...
    @name_str_id.setter
    def name_str_id(self, value: int) -> None: ...

    @property
    def creation_str_id(self) -> int: ...
    @creation_str_id.setter
    def creation_str_id(self, value: int) -> None: ...

    @property
    def help_str_id(self) -> int: ...
    @help_str_id.setter
    def help_str_id(self, value: int) -> None: ...

    @property
    def hotkey_text_str_id(self) -> int: ...
    @hotkey_text_str_id.setter
    def hotkey_text_str_id(self, value: int) -> None: ...

    @property
    def hotkey_str_id(self) -> int: ...
    @hotkey_str_id.setter
    def hotkey_str_id(self, value: int) -> None: ...

    @property
    def trait(self) -> int: ...
    @trait.setter
    def trait(self, value: int) -> None: ...

    @property
    def trait_piece(self) -> int: ...
    @trait_piece.setter
    def trait_piece(self, value: int) -> None: ...

    @property
    def standing_sprite_id1(self) -> int: ...
    @standing_sprite_id1.setter
    def standing_sprite_id1(self, value: int) -> None: ...

    @property
    def standing_sprite_id2(self) -> int: ...
    @standing_sprite_id2.setter
    def standing_sprite_id2(self, value: int) -> None: ...

    @property
    def dying_sprite_id(self) -> int: ...
    @dying_sprite_id.setter
    def dying_sprite_id(self, value: int) -> None: ...

    @property
    def undead_sprite_id(self) -> int: ...
    @undead_sprite_id.setter
    def undead_sprite_id(self, value: int) -> None: ...

    @property
    def icon_id(self) -> int: ...
    @icon_id.setter
    def icon_id(self, value: int) -> None: ...

    @property
    def radius_x(self) -> float: ...
    @radius_x.setter
    def radius_x(self, value: float) -> None: ...

    @property
    def radius_y(self) -> float: ...
    @radius_y.setter
    def radius_y(self, value: float) -> None: ...

    @property
    def radius_z(self) -> float: ...
    @radius_z.setter
    def radius_z(self, value: float) -> None: ...

    @property
    def selection_radius_x(self) -> float: ...
    @selection_radius_x.setter
    def selection_radius_x(self, value: float) -> None: ...

    @property
    def selection_radius_y(self) -> float: ...
    @selection_radius_y.setter
    def selection_radius_y(self, value: float) -> None: ...

    @property
    def selection_radius_z(self) -> float: ...
    @selection_radius_z.setter
    def selection_radius_z(self, value: float) -> None: ...

    @property
    def selection_effect(self) -> int: ...
    @selection_effect.setter
    def selection_effect(self, value: int) -> None: ...

    @property
    def editor_selection_color(self) -> int: ...
    @editor_selection_color.setter
    def editor_selection_color(self, value: int) -> None: ...

    @property
    def train_sound_id(self) -> int: ...
    @train_sound_id.setter
    def train_sound_id(self, value: int) -> None: ...

    @property
    def damage_sound_id1(self) -> int: ...
    @damage_sound_id1.setter
    def damage_sound_id1(self, value: int) -> None: ...

    @property
    def selection_sound_id(self) -> int: ...
    @selection_sound_id.setter
    def selection_sound_id(self, value: int) -> None: ...

    @property
    def dying_sound_id(self) -> int: ...
    @dying_sound_id.setter
    def dying_sound_id(self, value: int) -> None: ...

    @property
    def wwise_train_sound_id(self) -> int: ...
    @wwise_train_sound_id.setter
    def wwise_train_sound_id(self, value: int) -> None: ...

    @property
    def wwise_damage_sound_id(self) -> int: ...
    @wwise_damage_sound_id.setter
    def wwise_damage_sound_id(self, value: int) -> None: ...

    @property
    def wwise_selection_sound_id(self) -> int: ...
    @wwise_selection_sound_id.setter
    def wwise_selection_sound_id(self, value: int) -> None: ...

    @property
    def wwise_dying_sound_id(self) -> int: ...
    @wwise_dying_sound_id.setter
    def wwise_dying_sound_id(self, value: int) -> None: ...

    @property
    def dead_unit_id(self) -> int: ...
    @dead_unit_id.setter
    def dead_unit_id(self, value: int) -> None: ...

    @property
    def blood_unit_id(self) -> int: ...
    @blood_unit_id.setter
    def blood_unit_id(self, value: int) -> None: ...

    @property
    def undead_mode(self) -> int: ...
    @undead_mode.setter
    def undead_mode(self, value: int) -> None: ...

    @property
    def can_be_built_on(self) -> int: ...
    @can_be_built_on.setter
    def can_be_built_on(self, value: int) -> None: ...

    @property
    def required_side_terrain_id1(self) -> int: ...
    @required_side_terrain_id1.setter
    def required_side_terrain_id1(self, value: int) -> None: ...

    @property
    def required_side_terrain_id2(self) -> int: ...
    @required_side_terrain_id2.setter
    def required_side_terrain_id2(self, value: int) -> None: ...

    @property
    def required_center_terrain_id1(self) -> int: ...
    @required_center_terrain_id1.setter
    def required_center_terrain_id1(self, value: int) -> None: ...

    @property
    def required_center_terrain_id2(self) -> int: ...
    @required_center_terrain_id2.setter
    def required_center_terrain_id2(self, value: int) -> None: ...

    @property
    def required_clearance_radius_x(self) -> float: ...
    @required_clearance_radius_x.setter
    def required_clearance_radius_x(self, value: float) -> None: ...

    @property
    def required_clearance_radius_y(self) -> float: ...
    @required_clearance_radius_y.setter
    def required_clearance_radius_y(self, value: float) -> None: ...

    @property
    def elevation_restriction_mode(self) -> int: ...
    @elevation_restriction_mode.setter
    def elevation_restriction_mode(self, value: int) -> None: ...

    @property
    def terrain_restriction_id(self) -> int: ...
    @terrain_restriction_id.setter
    def terrain_restriction_id(self, value: int) -> None: ...

    @property
    def foundation_terrain_id(self) -> int: ...
    @foundation_terrain_id.setter
    def foundation_terrain_id(self, value: int) -> None: ...

    @property
    def movement_mode(self) -> int: ...
    @movement_mode.setter
    def movement_mode(self, value: int) -> None: ...

    @property
    def obstruction_type(self) -> int: ...
    @obstruction_type.setter
    def obstruction_type(self, value: int) -> None: ...

    @property
    def obstruction_class(self) -> int: ...
    @obstruction_class.setter
    def obstruction_class(self, value: int) -> None: ...

    @property
    def resource_carry_capacity(self) -> int: ...
    @resource_carry_capacity.setter
    def resource_carry_capacity(self, value: int) -> None: ...

    @property
    def resource_decay_rate(self) -> float: ...
    @resource_decay_rate.setter
    def resource_decay_rate(self, value: float) -> None: ...

    @property
    def resource_gather_group(self) -> int: ...
    @resource_gather_group.setter
    def resource_gather_group(self, value: int) -> None: ...

    @property
    def enable_auto_gather(self) -> int: ...
    @enable_auto_gather.setter
    def enable_auto_gather(self, value: int) -> None: ...

    @property
    def blast_defense_level(self) -> int: ...
    @blast_defense_level.setter
    def blast_defense_level(self, value: int) -> None: ...

    @property
    def combat_level(self) -> int: ...
    @combat_level.setter
    def combat_level(self, value: int) -> None: ...

    @property
    def old_attack_mode(self) -> int: ...
    @old_attack_mode.setter
    def old_attack_mode(self, value: int) -> None: ...

    @property
    def interaction_mode(self) -> int: ...
    @interaction_mode.setter
    def interaction_mode(self, value: int) -> None: ...

    @property
    def minimap_mode(self) -> int: ...
    @minimap_mode.setter
    def minimap_mode(self, value: int) -> None: ...

    @property
    def interface_mode(self) -> int: ...
    @interface_mode.setter
    def interface_mode(self, value: int) -> None: ...

    @property
    def minimap_color(self) -> int: ...
    @minimap_color.setter
    def minimap_color(self, value: int) -> None: ...

    @property
    def fog_visibility_mode(self) -> int: ...
    @fog_visibility_mode.setter
    def fog_visibility_mode(self, value: int) -> None: ...

    @property
    def occlusion_mode(self) -> int: ...
    @occlusion_mode.setter
    def occlusion_mode(self, value: int) -> None: ...

    @property
    def sort_number(self) -> int: ...
    @sort_number.setter
    def sort_number(self, value: int) -> None: ...

    @property
    def hide_in_editor(self) -> int: ...
    @hide_in_editor.setter
    def hide_in_editor(self, value: int) -> None: ...

    @property
    def multiple_attribute_mode(self) -> float: ...
    @multiple_attribute_mode.setter
    def multiple_attribute_mode(self, value: float) -> None: ...

    @property
    def recyclable(self) -> int: ...
    @recyclable.setter
    def recyclable(self, value: int) -> None: ...

    @property
    def doppelganger_mode(self) -> int: ...
    @doppelganger_mode.setter
    def doppelganger_mode(self, value: int) -> None: ...

    @property
    def convert_terrain(self) -> int: ...
    @convert_terrain.setter
    def convert_terrain(self, value: int) -> None: ...

    # --------------------------------------------------------------------------
    # Wrapper Accessors
    # --------------------------------------------------------------------------
    @property
    def combat(self) -> CombatWrapper: ...
    @property
    def creation(self) -> CreationWrapper: ...
    @property
    def cost(self) -> CostsManager: ...  # Wrapper points to CostsManager via property in unit_handle.py
    
    @property
    def movement(self) -> MovementWrapper: ...
    @property
    def behavior(self) -> BehaviorWrapper: ...
    @property
    def projectile(self) -> ProjectileWrapper: ...
    @property
    def building(self) -> BuildingWrapper: ...
    
    @property
    def resource_storages(self) -> ResourcesManager: ...
    @property
    def resources(self) -> ResourcesManager: ...
    
    @property
    def damage_graphics(self) -> DamageGraphicsManager: ...
    @property
    def train_locations_wrapper(self) -> TrainLocationsManager: ...
    
    @property
    def tasks(self) -> TasksManager: ...
    @property
    def tasks_wrapper(self) -> TasksManager: ...

    # --------------------------------------------------------------------------
    # Flattened Wrappers
    # --------------------------------------------------------------------------
    # BuildingWrapper
    @property
    def construction_graphic_id(self) -> int: ...
    @construction_graphic_id.setter
    def construction_graphic_id(self, value: int) -> None: ...

    @property
    def snow_graphic_id(self) -> int: ...
    @snow_graphic_id.setter
    def snow_graphic_id(self, value: int) -> None: ...

    @property
    def destruction_graphic_id(self) -> int: ...
    @destruction_graphic_id.setter
    def destruction_graphic_id(self, value: int) -> None: ...

    @property
    def destruction_rubble_graphic_id(self) -> int: ...
    @destruction_rubble_graphic_id.setter
    def destruction_rubble_graphic_id(self, value: int) -> None: ...

    @property
    def research_graphic_id(self) -> int: ...
    @research_graphic_id.setter
    def research_graphic_id(self, value: int) -> None: ...

    @property
    def research_complete_graphic_id(self) -> int: ...
    @research_complete_graphic_id.setter
    def research_complete_graphic_id(self, value: int) -> None: ...

    @property
    def adjacent_mode(self) -> int: ...
    @adjacent_mode.setter
    def adjacent_mode(self, value: int) -> None: ...

    @property
    def graphics_angle(self) -> int: ...
    @graphics_angle.setter
    def graphics_angle(self, value: int) -> None: ...

    @property
    def disappears_when_built(self) -> int: ...
    @disappears_when_built.setter
    def disappears_when_built(self, value: int) -> None: ...

    @property
    def stack_unit_id(self) -> int: ...
    @stack_unit_id.setter
    def stack_unit_id(self, value: int) -> None: ...

    @property
    def old_overlap_id(self) -> int: ...
    @old_overlap_id.setter
    def old_overlap_id(self, value: int) -> None: ...

    @property
    def tech_id(self) -> int: ...
    @tech_id.setter
    def tech_id(self, value: int) -> None: ...

    @property
    def completion_tech_id(self) -> int: ...
    @completion_tech_id.setter
    def completion_tech_id(self, value: int) -> None: ...

    @property
    def can_burn(self) -> int: ...
    @can_burn.setter
    def can_burn(self, value: int) -> None: ...

    @property
    def head_unit_id(self) -> int: ...
    @head_unit_id.setter
    def head_unit_id(self, value: int) -> None: ...

    @property
    def transform_unit_id(self) -> int: ...
    @transform_unit_id.setter
    def transform_unit_id(self, value: int) -> None: ...

    @property
    def salvage_unit_id(self) -> int: ...
    @salvage_unit_id.setter
    def salvage_unit_id(self, value: int) -> None: ...

    @property
    def pile_unit_id(self) -> int: ...
    @pile_unit_id.setter
    def pile_unit_id(self, value: int) -> None: ...

    @property
    def transform_sound_id(self) -> int: ...
    @transform_sound_id.setter
    def transform_sound_id(self, value: int) -> None: ...

    @property
    def construction_sound_id(self) -> int: ...
    @construction_sound_id.setter
    def construction_sound_id(self, value: int) -> None: ...

    @property
    def wwise_transform_sound_id(self) -> int: ...
    @wwise_transform_sound_id.setter
    def wwise_transform_sound_id(self, value: int) -> None: ...

    @property
    def wwise_construction_sound_id(self) -> int: ...
    @wwise_construction_sound_id.setter
    def wwise_construction_sound_id(self, value: int) -> None: ...

    @property
    def garrison_type(self) -> int: ...
    @garrison_type.setter
    def garrison_type(self, value: int) -> None: ...

    @property
    def garrison_heal_rate(self) -> float: ...
    @garrison_heal_rate.setter
    def garrison_heal_rate(self, value: float) -> None: ...

    @property
    def garrison_repair_rate(self) -> float: ...
    @garrison_repair_rate.setter
    def garrison_repair_rate(self, value: float) -> None: ...

    @property
    def looting_table(self) -> Any: ...
    @looting_table.setter
    def looting_table(self, value: Any) -> None: ...

    @property
    def annexes(self) -> AnnexesManager: ...
    @annexes.setter
    def annexes(self, value: AnnexesManager) -> None: ...

    @property
    def annexes_manager(self) -> AnnexesManager: ...
    @annexes_manager.setter
    def annexes_manager(self, value: AnnexesManager) -> None: ...

    # CombatWrapper
    @property
    def attack_graphic_id(self) -> int: ...
    @attack_graphic_id.setter
    def attack_graphic_id(self, value: int) -> None: ...

    @property
    def attack_graphic_2_id(self) -> int: ...
    @attack_graphic_2_id.setter
    def attack_graphic_2_id(self, value: int) -> None: ...

    @property
    def max_range(self) -> float: ...
    @max_range.setter
    def max_range(self, value: float) -> None: ...

    @property
    def min_range(self) -> float: ...
    @min_range.setter
    def min_range(self, value: float) -> None: ...

    @property
    def reload_time(self) -> float: ...
    @reload_time.setter
    def reload_time(self, value: float) -> None: ...

    @property
    def accuracy_percent(self) -> int: ...
    @accuracy_percent.setter
    def accuracy_percent(self, value: int) -> None: ...

    @property
    def accuracy_dispersion(self) -> float: ...
    @accuracy_dispersion.setter
    def accuracy_dispersion(self, value: float) -> None: ...

    @property
    def blast_width(self) -> float: ...
    @blast_width.setter
    def blast_width(self, value: float) -> None: ...

    @property
    def blast_damage(self) -> float: ...
    @blast_damage.setter
    def blast_damage(self, value: float) -> None: ...

    @property
    def blast_attack_level(self) -> int: ...
    @blast_attack_level.setter
    def blast_attack_level(self, value: int) -> None: ...

    @property
    def frame_delay(self) -> int: ...
    @frame_delay.setter
    def frame_delay(self, value: int) -> None: ...

    @property
    def break_off_combat(self) -> int: ...
    @break_off_combat.setter
    def break_off_combat(self, value: int) -> None: ...

    @property
    def base_armor(self) -> int: ...
    @base_armor.setter
    def base_armor(self, value: int) -> None: ...

    @property
    def defense_terrain_bonus(self) -> int: ...
    @defense_terrain_bonus.setter
    def defense_terrain_bonus(self, value: int) -> None: ...

    @property
    def bonus_damage_resistance(self) -> float: ...
    @bonus_damage_resistance.setter
    def bonus_damage_resistance(self, value: float) -> None: ...

    @property
    def damage_reflection(self) -> float: ...
    @damage_reflection.setter
    def damage_reflection(self, value: float) -> None: ...

    @property
    def friendly_fire_damage(self) -> float: ...
    @friendly_fire_damage.setter
    def friendly_fire_damage(self, value: float) -> None: ...

    @property
    def displayed_attack(self) -> int: ...
    @displayed_attack.setter
    def displayed_attack(self, value: int) -> None: ...

    @property
    def displayed_melee_armour(self) -> int: ...
    @displayed_melee_armour.setter
    def displayed_melee_armour(self, value: int) -> None: ...

    @property
    def displayed_range(self) -> float: ...
    @displayed_range.setter
    def displayed_range(self, value: float) -> None: ...

    @property
    def displayed_reload_time(self) -> float: ...
    @displayed_reload_time.setter
    def displayed_reload_time(self, value: float) -> None: ...

    @property
    def projectile_unit_id(self) -> int: ...
    @projectile_unit_id.setter
    def projectile_unit_id(self, value: int) -> None: ...

    @property
    def graphic_displacement(self) -> Tuple[float, float, float]: ...
    @graphic_displacement.setter
    def graphic_displacement(self, value: Tuple[float, float, float]) -> None: ...

    @property
    def weapon_offset_x(self) -> float: ...
    @weapon_offset_x.setter
    def weapon_offset_x(self, value: float) -> None: ...

    @property
    def weapon_offset_y(self) -> float: ...
    @weapon_offset_y.setter
    def weapon_offset_y(self, value: float) -> None: ...

    @property
    def weapon_offset_z(self) -> float: ...
    @weapon_offset_z.setter
    def weapon_offset_z(self, value: float) -> None: ...

    @property
    def interrupt_frame(self) -> int: ...
    @interrupt_frame.setter
    def interrupt_frame(self, value: int) -> None: ...

    @property
    def garrison_firepower(self) -> float: ...
    @garrison_firepower.setter
    def garrison_firepower(self, value: float) -> None: ...

    # CreationWrapper
    @property
    def train_time(self) -> int: ...
    @train_time.setter
    def train_time(self, value: int) -> None: ...

    @property
    def train_location_id(self) -> int: ...
    @train_location_id.setter
    def train_location_id(self, value: int) -> None: ...

    @property
    def button_id(self) -> int: ...
    @button_id.setter
    def button_id(self, value: int) -> None: ...

    @property
    def hot_key_id(self) -> int: ...
    @hot_key_id.setter
    def hot_key_id(self, value: int) -> None: ...

    @property
    def garrison_graphic_id(self) -> int: ...
    @garrison_graphic_id.setter
    def garrison_graphic_id(self, value: int) -> None: ...

    @property
    def spawning_graphic_id(self) -> int: ...
    @spawning_graphic_id.setter
    def spawning_graphic_id(self, value: int) -> None: ...

    @property
    def upgrade_graphic_id(self) -> int: ...
    @upgrade_graphic_id.setter
    def upgrade_graphic_id(self, value: int) -> None: ...

    @property
    def hero_glow_graphic_id(self) -> int: ...
    @hero_glow_graphic_id.setter
    def hero_glow_graphic_id(self, value: int) -> None: ...

    @property
    def idle_attack_graphic_id(self) -> int: ...
    @idle_attack_graphic_id.setter
    def idle_attack_graphic_id(self, value: int) -> None: ...

    @property
    def special_graphic_id(self) -> int: ...
    @special_graphic_id.setter
    def special_graphic_id(self, value: int) -> None: ...

    @property
    def max_charge(self) -> float: ...
    @max_charge.setter
    def max_charge(self, value: float) -> None: ...

    @property
    def recharge_rate(self) -> float: ...
    @recharge_rate.setter
    def recharge_rate(self, value: float) -> None: ...

    @property
    def charge_event(self) -> int: ...
    @charge_event.setter
    def charge_event(self, value: int) -> None: ...

    @property
    def charge_type(self) -> int: ...
    @charge_type.setter
    def charge_type(self, value: int) -> None: ...

    @property
    def charge_target(self) -> int: ...
    @charge_target.setter
    def charge_target(self, value: int) -> None: ...

    @property
    def charge_projectile_unit_id(self) -> int: ...
    @charge_projectile_unit_id.setter
    def charge_projectile_unit_id(self, value: int) -> None: ...

    @property
    def rear_attack_modifier(self) -> float: ...
    @rear_attack_modifier.setter
    def rear_attack_modifier(self, value: float) -> None: ...

    @property
    def flank_attack_modifier(self) -> float: ...
    @flank_attack_modifier.setter
    def flank_attack_modifier(self, value: float) -> None: ...

    @property
    def attack_priority(self) -> int: ...
    @attack_priority.setter
    def attack_priority(self, value: int) -> None: ...

    @property
    def invulnerability_level(self) -> float: ...
    @invulnerability_level.setter
    def invulnerability_level(self, value: float) -> None: ...

    @property
    def min_conversion_time_mod(self) -> float: ...
    @min_conversion_time_mod.setter
    def min_conversion_time_mod(self, value: float) -> None: ...

    @property
    def max_conversion_time_mod(self) -> float: ...
    @max_conversion_time_mod.setter
    def max_conversion_time_mod(self, value: float) -> None: ...

    @property
    def conversion_chance_mod(self) -> float: ...
    @conversion_chance_mod.setter
    def conversion_chance_mod(self, value: float) -> None: ...

    @property
    def total_projectiles(self) -> float: ...
    @total_projectiles.setter
    def total_projectiles(self, value: float) -> None: ...

    @property
    def max_total_projectiles(self) -> int: ...
    @max_total_projectiles.setter
    def max_total_projectiles(self, value: int) -> None: ...

    @property
    def secondary_projectile_unit_id(self) -> int: ...
    @secondary_projectile_unit_id.setter
    def secondary_projectile_unit_id(self, value: int) -> None: ...

    @property
    def projectile_spawning_area(self) -> Tuple[float, float, float]: ...
    @projectile_spawning_area.setter
    def projectile_spawning_area(self, value: Tuple[float, float, float]) -> None: ...

    @property
    def projectile_spawning_area_width(self) -> float: ...
    @projectile_spawning_area_width.setter
    def projectile_spawning_area_width(self, value: float) -> None: ...

    @property
    def projectile_spawning_area_length(self) -> float: ...
    @projectile_spawning_area_length.setter
    def projectile_spawning_area_length(self, value: float) -> None: ...

    @property
    def projectile_spawning_area_randomness(self) -> float: ...
    @projectile_spawning_area_randomness.setter
    def projectile_spawning_area_randomness(self, value: float) -> None: ...

    @property
    def button_icon_id(self) -> int: ...
    @button_icon_id.setter
    def button_icon_id(self, value: int) -> None: ...

    @property
    def button_short_tooltip_id(self) -> int: ...
    @button_short_tooltip_id.setter
    def button_short_tooltip_id(self, value: int) -> None: ...

    @property
    def button_extended_tooltip_id(self) -> int: ...
    @button_extended_tooltip_id.setter
    def button_extended_tooltip_id(self, value: int) -> None: ...

    @property
    def button_hotkey_action(self) -> int: ...
    @button_hotkey_action.setter
    def button_hotkey_action(self, value: int) -> None: ...

    @property
    def creatable_type(self) -> int: ...
    @creatable_type.setter
    def creatable_type(self, value: int) -> None: ...

    @property
    def hero_mode(self) -> int: ...
    @hero_mode.setter
    def hero_mode(self, value: int) -> None: ...

    @property
    def special_ability(self) -> int: ...
    @special_ability.setter
    def special_ability(self, value: int) -> None: ...

    @property
    def displayed_pierce_armour(self) -> int: ...
    @displayed_pierce_armour.setter
    def displayed_pierce_armour(self, value: int) -> None: ...

    # MovementWrapper
    @property
    def walking_graphic_id(self) -> int: ...
    @walking_graphic_id.setter
    def walking_graphic_id(self, value: int) -> None: ...

    @property
    def running_graphic_id(self) -> int: ...
    @running_graphic_id.setter
    def running_graphic_id(self, value: int) -> None: ...

    @property
    def rotation_speed(self) -> float: ...
    @rotation_speed.setter
    def rotation_speed(self, value: float) -> None: ...

    @property
    def turn_radius(self) -> float: ...
    @turn_radius.setter
    def turn_radius(self, value: float) -> None: ...

    @property
    def rotation_radius(self) -> float: ...
    @rotation_radius.setter
    def rotation_radius(self, value: float) -> None: ...

    @property
    def turn_radius_speed(self) -> float: ...
    @turn_radius_speed.setter
    def turn_radius_speed(self, value: float) -> None: ...

    @property
    def rotation_radius_speed(self) -> float: ...
    @rotation_radius_speed.setter
    def rotation_radius_speed(self, value: float) -> None: ...

    @property
    def max_yaw_per_second_moving(self) -> float: ...
    @max_yaw_per_second_moving.setter
    def max_yaw_per_second_moving(self, value: float) -> None: ...

    @property
    def max_yaw_per_sec_walking(self) -> float: ...
    @max_yaw_per_sec_walking.setter
    def max_yaw_per_sec_walking(self, value: float) -> None: ...

    @property
    def stationary_yaw_revolution_time(self) -> float: ...
    @stationary_yaw_revolution_time.setter
    def stationary_yaw_revolution_time(self, value: float) -> None: ...

    @property
    def standing_yaw_revolution_time(self) -> float: ...
    @standing_yaw_revolution_time.setter
    def standing_yaw_revolution_time(self, value: float) -> None: ...

    @property
    def max_yaw_per_second_stationary(self) -> float: ...
    @max_yaw_per_second_stationary.setter
    def max_yaw_per_second_stationary(self, value: float) -> None: ...

    @property
    def max_yaw_per_sec_standing(self) -> float: ...
    @max_yaw_per_sec_standing.setter
    def max_yaw_per_sec_standing(self, value: float) -> None: ...

    @property
    def tracking_unit_id(self) -> int: ...
    @tracking_unit_id.setter
    def tracking_unit_id(self, value: int) -> None: ...

    @property
    def trailing_unit_id(self) -> int: ...
    @trailing_unit_id.setter
    def trailing_unit_id(self, value: int) -> None: ...

    @property
    def tracking_unit_mode(self) -> int: ...
    @tracking_unit_mode.setter
    def tracking_unit_mode(self, value: int) -> None: ...

    @property
    def trail_mode(self) -> int: ...
    @trail_mode.setter
    def trail_mode(self, value: int) -> None: ...

    @property
    def tracking_unit_density(self) -> float: ...
    @tracking_unit_density.setter
    def tracking_unit_density(self, value: float) -> None: ...

    @property
    def trail_spacing(self) -> float: ...
    @trail_spacing.setter
    def trail_spacing(self, value: float) -> None: ...

    @property
    def old_size_class(self) -> int: ...
    @old_size_class.setter
    def old_size_class(self, value: int) -> None: ...

    @property
    def old_move_algorithm(self) -> int: ...
    @old_move_algorithm.setter
    def old_move_algorithm(self, value: int) -> None: ...

    @property
    def min_collision_size_multiplier(self) -> float: ...
    @min_collision_size_multiplier.setter
    def min_collision_size_multiplier(self, value: float) -> None: ...

    # BehaviorWrapper
    @property
    def default_task_id(self) -> int: ...
    @default_task_id.setter
    def default_task_id(self, value: int) -> None: ...

    @property
    def search_radius(self) -> float: ...
    @search_radius.setter
    def search_radius(self, value: float) -> None: ...

    @property
    def work_rate(self) -> float: ...
    @work_rate.setter
    def work_rate(self, value: float) -> None: ...

    @property
    def task_swap_group(self) -> int: ...
    @task_swap_group.setter
    def task_swap_group(self, value: int) -> None: ...

    @property
    def run_mode(self) -> int: ...
    @run_mode.setter
    def run_mode(self, value: int) -> None: ...

    @property
    def run_pattern(self) -> int: ...
    @run_pattern.setter
    def run_pattern(self, value: int) -> None: ...

    @property
    def drop_site_unit_ids(self) -> DropSitesManager: ...
    @drop_site_unit_ids.setter
    def drop_site_unit_ids(self, value: DropSitesManager) -> None: ...

    @property
    def attack_sound_id(self) -> int: ...
    @attack_sound_id.setter
    def attack_sound_id(self, value: int) -> None: ...

    @property
    def attack_sound(self) -> int: ...
    @attack_sound.setter
    def attack_sound(self, value: int) -> None: ...

    @property
    def move_sound_id(self) -> int: ...
    @move_sound_id.setter
    def move_sound_id(self, value: int) -> None: ...

    @property
    def move_sound(self) -> int: ...
    @move_sound.setter
    def move_sound(self, value: int) -> None: ...

    @property
    def wwise_attack_sound_id(self) -> int: ...
    @wwise_attack_sound_id.setter
    def wwise_attack_sound_id(self, value: int) -> None: ...

    @property
    def wwise_move_sound_id(self) -> int: ...
    @wwise_move_sound_id.setter
    def wwise_move_sound_id(self, value: int) -> None: ...

    # ProjectileWrapper
    @property
    def projectile_type(self) -> int: ...
    @projectile_type.setter
    def projectile_type(self, value: int) -> None: ...

    @property
    def smart_mode(self) -> int: ...
    @smart_mode.setter
    def smart_mode(self, value: int) -> None: ...

    @property
    def hit_mode(self) -> int: ...
    @hit_mode.setter
    def hit_mode(self, value: int) -> None: ...

    @property
    def vanish_mode(self) -> int: ...
    @vanish_mode.setter
    def vanish_mode(self, value: int) -> None: ...

    @property
    def area_effect_specials(self) -> int: ...
    @area_effect_specials.setter
    def area_effect_specials(self, value: int) -> None: ...

    @property
    def projectile_arc(self) -> float: ...
    @projectile_arc.setter
    def projectile_arc(self, value: float) -> None: ...

    # --------------------------------------------------------------------------
    # Collection Accessors (Flattened)
    # --------------------------------------------------------------------------
    @property
    def attacks(self) -> AttacksManager: ...
    @attacks.setter
    def attacks(self, value: List[Any]) -> None: ...
    
    @property
    def armours(self) -> ArmoursManager: ...
    @armours.setter
    def armours(self, value: List[Any]) -> None: ...
    
    @property
    def resource_costs(self) -> CostsManager: ...
    @resource_costs.setter
    def resource_costs(self, value: List[Any]) -> None: ...
    
    @property
    def costs(self) -> CostsManager: ...
    @costs.setter
    def costs(self, value: List[Any]) -> None: ...
    
    @property
    def train_locations(self) -> TrainLocationsManager: ...
    @train_locations.setter
    def train_locations(self, value: List[Any]) -> None: ...
    
    @property
    def drop_sites(self) -> DropSitesManager: ...
    @drop_sites.setter
    def drop_sites(self, value: List[int]) -> None: ...
    
    # --------------------------------------------------------------------------
    # Methods
    # --------------------------------------------------------------------------
    def __init__(self, workspace: GenieWorkspace, unit_id: int, civ_ids: Optional[List[int]] = None) -> None: ...
    
    def invalidate_cache(self) -> None: ...
    
    # Resource Storage
    def resource_1(self, type: int = 0, amount: float = 0.0, flag: int = 0) -> None: ...
    def resource_2(self, type: int = 0, amount: float = 0.0, flag: int = 0) -> None: ...
    def resource_3(self, type: int = 0, amount: float = 0.0, flag: int = 0) -> None: ...
    
    # Attacks
    def add_attack(self, class_: int, amount: int) -> Optional[AttackHandle]: ...
    def get_attack_by_id(self, attack_id: int) -> Optional[AttackHandle]: ...
    def get_attack_by_class(self, class_: int) -> Optional[AttackHandle]: ...
    def remove_attack(self, attack_id: int) -> bool: ...
    def remove_all_attacks(self) -> None: ...
    def set_attack(self, class_: int, amount: int) -> Optional[AttackHandle]: ...
    
    # Armours
    def add_armour(self, class_: int, amount: int) -> Optional[ArmourHandle]: ...
    def get_armour_by_id(self, armour_id: int) -> Optional[ArmourHandle]: ...
    def get_armour_by_class(self, class_: int) -> Optional[ArmourHandle]: ...
    def remove_armour(self, armour_id: int) -> bool: ...
    def set_armour(self, class_: int, amount: int) -> Optional[ArmourHandle]: ...
    
    # Damage Graphics
    def add_damage_graphic(self, graphic_id: int, damage_percent: int, apply_mode: int = 0) -> Optional[DamageGraphicHandle]: ...
    def get_damage_graphic(self, damage_graphic_id: int) -> Optional[DamageGraphicHandle]: ...
    def remove_damage_graphic(self, damage_graphic_id: int) -> bool: ...
    def remove_all_damage_graphics(self) -> None: ...
    # Tasks
    @property
    def add_task(self) -> TaskBuilder: ...
    
    def create_task(
        self,
        task_type: int = 0,
        id: int = 0,
        is_default: int = 0,
        action_type: int = 0,
        class_id: int = -1,
        unit_id: int = -1,
        terrain_id: int = -1,
        resource_in: int = -1,
        resource_out: int = -1,
        work_value_1: float = 0.0,
        work_value_2: float = 0.0,
        work_range: float = 0.0,
        enabled: int = 1,
        **kwargs
    ) -> Optional[TaskHandle]: ...
    
    def get_task(self, task_id: int) -> Optional[TaskHandle]: ...
    def get_tasks_list(self) -> List[TaskHandle]: ...
    def remove_task(self, task_id: int) -> bool: ...
    
    # Train Locations
    def add_train_location(
        self,
        unit_id: int,
        train_time: int = 0,
        button_id: int = 0,
        hot_key_id: int = 0,
        **kwargs
    ) -> Optional[TrainLocationHandle]: ...
    
    def get_train_location(self, train_location_id: int) -> Optional[TrainLocationHandle]: ...
    def remove_train_location(self, train_location_id: int) -> bool: ...
    
    # Drop Sites
    def add_drop_site(self, unit_id: int) -> Optional[DropSiteHandle]: ...
    def get_drop_site(self, drop_site_id: int) -> Optional[DropSiteHandle]: ...
    def remove_drop_site(self, drop_site_id: int) -> bool: ...



"""
GraphicHandle - Wrapper for individual Sprite objects.

Provides attribute access to sprite properties.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Optional, Union

from aoe2_genie_tooling.Base.core.typed_ids import GraphicId, DeltaIndex

if TYPE_CHECKING:
    from aoe2_genie_tooling.Base.workspace import GenieWorkspace
    from aoe2_genie_tooling.Graphics.delta_handle import DeltaHandle

__all__ = ["GraphicHandle"]


class GraphicHandle:
    """
    Handle for a single sprite/graphic.
    
    Provides direct attribute access to the underlying Sprite object.
    """
    
    def __init__(self, workspace: GenieWorkspace, graphic_id: int) -> None:
        """
        Initialize GraphicHandle.
        
        Args:
            workspace: The GenieWorkspace instance
            graphic_id: ID of the graphic
        """
        object.__setattr__(self, '_workspace', workspace)
        object.__setattr__(self, '_id', graphic_id)
        object.__setattr__(self, '_sprite', workspace.dat.sprites[graphic_id])
    
    @property
    def id(self) -> GraphicId:
        """Get the graphic ID (typed for IDE checking)."""
        return GraphicId(self._id)
    
    @property
    def workspace(self) -> GenieWorkspace:
        """Get the workspace."""
        return self._workspace
    
    def __getattr__(self, name: str) -> Any:
        """
        Get attribute from underlying sprite.
        
        Args:
            name: Attribute name
            
        Returns:
            Attribute value from sprite
        """
        return getattr(self._sprite, name)
    
    def __setattr__(self, name: str, value: Any) -> None:
        """
        Set attribute on underlying sprite.
        
        Args:
            name: Attribute name
            value: Value to set
        """
        if name.startswith('_'):
            object.__setattr__(self, name, value)
        # Check if this class has a property descriptor for the name
        elif hasattr(type(self), name) and isinstance(getattr(type(self), name), property):
            prop = getattr(type(self), name)
            if prop.fset is not None:
                prop.fset(self, value)
            else:
                raise AttributeError(f"property '{name}' has no setter")
        else:
            setattr(self._sprite, name, value)
    
    def __repr__(self) -> str:
        return f"GraphicHandle(id={self._id})"
    
    # =========================================================================
    # PROPERTY ALIASES (Match Documentation)
    # =========================================================================
    
    @property
    def slp(self) -> int:
        """Alias for slp_id."""
        return self._sprite.slp_id
    
    @slp.setter
    def slp(self, value: int) -> None:
        self._sprite.slp_id = value
        
    @property
    def frame_count(self) -> int:
        """Alias for num_frames."""
        return self._sprite.num_frames
    
    @frame_count.setter
    def frame_count(self, value: int) -> None:
        self._sprite.num_frames = value
        
    @property
    def angle_count(self) -> int:
        """Alias for num_facets."""
        return self._sprite.num_facets
    
    @angle_count.setter
    def angle_count(self, value: int) -> None:
        self._sprite.num_facets = value
        
    @property
    def frame_duration(self) -> float:
        """Alias for frame_rate (per-frame duration)."""
        return self._sprite.frame_rate
    
    @frame_duration.setter
    def frame_duration(self, value: float) -> None:
        self._sprite.frame_rate = value
    
    @property
    def animation_duration(self) -> float:
        """Total animation duration (frame_count * frame_duration)."""
        return self._sprite.num_frames * self._sprite.frame_rate
    
    @animation_duration.setter
    def animation_duration(self, value: float) -> None:
        """Set total duration by calculating frame_rate from frame_count."""
        if self._sprite.num_frames > 0:
            self._sprite.frame_rate = value / self._sprite.num_frames
        else:
            self._sprite.frame_rate = value  # Fallback if no frames
        
    @property
    def speed_multiplier(self) -> float:
        """Alias for speed_mult."""
        return self._sprite.speed_mult
    
    @speed_multiplier.setter
    def speed_multiplier(self, value: float) -> None:
        self._sprite.speed_mult = value
        
    @property
    def angle_sounds_used(self) -> bool:
        """Alias for facets_have_attack_sounds."""
        return self._sprite.facets_have_attack_sounds
    
    @angle_sounds_used.setter
    def angle_sounds_used(self, value: bool) -> None:
        self._sprite.facets_have_attack_sounds = value
        
    @property
    def player_color(self) -> int:
        """Alias for force_player_color."""
        return self._sprite.force_player_color
    
    @player_color.setter
    def player_color(self, value: int) -> None:
        self._sprite.force_player_color = value
        
    @property
    def coordinates(self) -> tuple[int, int, int, int]:
        """Alias for bounding_box (returning as tuple)."""
        return tuple(self._sprite.bounding_box) # type: ignore
    
    @coordinates.setter
    def coordinates(self, value: list[int] | tuple[int, int, int, int]) -> None:
        self._sprite.bounding_box = list(value)
        
    @property
    def editor_flag(self) -> int:
        """Alias for editor_mode."""
        return self._sprite.editor_mode
    
    @editor_flag.setter
    def editor_flag(self, value: int) -> None:
        self._sprite.editor_mode = value
    
    # =========================================================================
    # DELTA MANAGEMENT
    # =========================================================================
    
    def add_delta(
        self,
        graphic_id: Union[int, str, Any],
        offset_x: int = 0,
        offset_y: int = 0,
        display_angle: int = -1,
    ) -> DeltaHandle:
        """
        Add a delta (sub-graphic) to this graphic.
        
        Args:
            graphic_id: ID of graphic to attach as delta (int, UUID, or Handle)
            offset_x: X offset from parent (default: 0)
            offset_y: Y offset from parent (default: 0)
            display_angle: Angle filter, -1 = all angles (default: -1)
            
        Returns:
            DeltaHandle for the new delta
            
        Example:
            >>> gfx = gm.get(100)
            >>> shadow = gfx.add_delta(graphic_id=200, offset_y=5)
            >>> shadow.offset_x = 2
        """
        from sections.sprite_data.sprite_delta import SpriteDelta
        from aoe2_genie_tooling.Graphics.delta_handle import DeltaHandle
        
        # Resolve graphic_id (accepts int, GraphicHandle, or UUID - NOT DeltaHandle!)
        resolved_id = self._workspace.validator.resolve_id(
            graphic_id,
            "graphics",
            self._workspace,
            expected_handle_type=GraphicHandle,
            param_name="graphic_id",
        )
        
        # Create new delta with same version as parent
        new_delta = SpriteDelta(ver=self._sprite.ver)
        new_delta.sprite_id = resolved_id
        new_delta.offset_x = offset_x
        new_delta.offset_y = offset_y
        new_delta.display_angle = display_angle
        
        # CRITICAL: Don't use append()! bfp_rs lists share internal storage.
        current_deltas = list(self._sprite.deltas)
        current_deltas.append(new_delta)
        self._sprite.deltas = current_deltas  # setattr triggers bfp_rs copy
        
        # Update count
        self._sprite.num_deltas = len(self._sprite.deltas)
        
        # Return handle for new delta
        return DeltaHandle(self, len(self._sprite.deltas) - 1)
    
    def get_delta(self, delta_id: int) -> Optional[DeltaHandle]:
        """
        Get a delta by index.
        
        Args:
            delta_id: Index in deltas list (0-based)
            
        Returns:
            DeltaHandle if exists, None otherwise
        """
        from aoe2_genie_tooling.Graphics.delta_handle import DeltaHandle
        
        if 0 <= delta_id < len(self._sprite.deltas):
            return DeltaHandle(self, delta_id)
        return None
    
    def remove_delta(self, delta_id: int) -> bool:
        """
        Remove a delta by index.
        
        Args:
            delta_id: Index in deltas list (0-based)
            
        Returns:
            True if removed, False if index invalid
        """
        if 0 <= delta_id < len(self._sprite.deltas):
            del self._sprite.deltas[delta_id]
            self._sprite.num_deltas = len(self._sprite.deltas)
            return True
        return False
    
    def add_delta_from_graphic(
        self,
        graphic_id: Union[int, str, Any],
        delta_id: int,
    ) -> DeltaHandle:
        """
        Copy a delta from another graphic and add it to this graphic.
        
        Args:
            graphic_id: Source graphic ID (int, UUID, or Handle)
            delta_id: Delta index in source graphic
            
        Returns:
            DeltaHandle for the new delta
            
        Raises:
            InvalidIdError: If graphic_id or delta_id is invalid
            
        Example:
            >>> new_gfx.add_delta_from_graphic(graphic_id=100, delta_id=0)
        """
        from aoe2_genie_tooling.Graphics.delta_handle import DeltaHandle
        from aoe2_genie_tooling.Base.core.exceptions import InvalidIdError
        from aoe2_genie_tooling.Base.core.typed_ids import GraphicId, DeltaIndex
        
        validator = self._workspace.validator
        
        # Validate delta_id type - should be int, DeltaIndex, or DeltaHandle, NOT GraphicHandle/GraphicId
        resolved_delta_id = delta_id
        if hasattr(delta_id, 'delta_id'):
            # It's a DeltaHandle - extract the index
            resolved_delta_id = int(delta_id.delta_id)
        elif hasattr(delta_id, 'id'):
            # It's some other Handle (like GraphicHandle) - ERROR!
            actual_type = type(delta_id).__name__
            raise InvalidIdError(
                f"Expected int or DeltaHandle, got {actual_type}",
                action_description=f"Wrong type passed to 'delta_id'",
                context=f"Parameter: delta_id",
                hints=[
                    f"You passed a {actual_type} where an int or DeltaHandle was expected",
                    f"delta_id should be an index (0, 1, 2...) or a DeltaHandle object",
                    f"If you meant to pass a graphic, use the 'graphic_id' parameter instead",
                ],
            )
        elif isinstance(delta_id, GraphicId):
            # GraphicId passed to delta_id - ERROR!
            raise InvalidIdError(
                f"Expected int or DeltaIndex, got GraphicId",
                action_description=f"Wrong ID type passed to 'delta_id'",
                context=f"Parameter: delta_id",
                hints=[
                    f"You passed a GraphicId where a delta index was expected",
                    f"This usually means you're passing graphic.id to a delta_id parameter",
                    f"delta_id should be an index (0, 1, 2...) not a graphic ID",
                ],
            )
        elif isinstance(delta_id, DeltaIndex):
            resolved_delta_id = int(delta_id)
        elif not isinstance(delta_id, int):
            raise InvalidIdError(
                f"delta_id must be int or DeltaHandle, got {type(delta_id).__name__}",
                action_description=f"Invalid type for 'delta_id'",
            )
        
        # Resolve source graphic ID (must be GraphicHandle, not DeltaHandle!)
        resolved_gfx_id = validator.resolve_id(
            graphic_id,
            "graphics",
            self._workspace,
            expected_handle_type=GraphicHandle,
            param_name="graphic_id",
        )
        
        # Get source graphic
        source_gfx = self._workspace.graphic_manager.get(resolved_gfx_id)
        
        # Build rich context for error message
        gfx_name = source_gfx.name if source_gfx else "Unknown"
        context = f"Graphic '{gfx_name}' (ID: {resolved_gfx_id})"
        
        # Build current items list
        current_items = []
        if source_gfx:
            for i, delta in enumerate(source_gfx.deltas):
                current_items.append(f"Delta {i}: → Graphic {delta.graphic_id}")
        if not current_items:
            current_items = ["(no deltas exist)"]
        
        # Validate delta index with rich context
        validator.validate_index(
            resolved_delta_id,
            source_gfx._sprite.deltas,
            "delta_id",
            context=context,
            current_items=current_items,
            hints=[
                f"This graphic has {len(source_gfx.deltas)} delta(s)",
                "Use source_graphic.deltas to list all deltas",
                "Use source_graphic.add_delta() to add a new delta first",
            ],
            action_description="Attempt to copy a delta that doesn't exist",
        )
        
        # Get source delta
        source_delta = source_gfx.deltas[resolved_delta_id]
        
        # Create new delta with same properties
        return self.add_delta(
            graphic_id=source_delta.graphic_id,
            offset_x=source_delta.offset_x,
            offset_y=source_delta.offset_y,
            display_angle=source_delta.display_angle,
        )
    
    def clear_deltas(self) -> None:
        """Remove all deltas from this graphic."""
        self._sprite.deltas.clear()
        self._sprite.num_deltas = 0

    def exists(self) -> bool:
        """
        Check if this graphic entry exists and is not None.
        
        Returns:
            True if graphic exists
        """
        return self._sprite is not None
    
    def add_angle_sound(
        self,
        frame_num: int = 0,
        sound_id: int = -1,
        wwise_sound_id: int = 0,
        frame_num_2: int = 0,
        sound_id_2: int = -1,
        wwise_sound_id_2: int = 0,
        frame_num_3: int = 0,
        sound_id_3: int = -1,
        wwise_sound_id_3: int = 0,
    ) -> None:
        """
        Add an angle sound entry.
        
        Args:
            frame_num: Trigger frame for sound 1
            sound_id: Sound ID 1
            wwise_sound_id: Wwise sound ID 1 (DE2)
            frame_num_2: Trigger frame for sound 2
            sound_id_2: Sound ID 2
            wwise_sound_id_2: Wwise sound ID 2 (DE2)
            frame_num_3: Trigger frame for sound 3
            sound_id_3: Sound ID 3
            wwise_sound_id_3: Wwise sound ID 3 (DE2)
        """
        from sections.sprite_data.facet_attack_sound import FacetAttackSound
        
        angle_sound = FacetAttackSound(ver=self._sprite.ver)
        angle_sound.sound_delay1 = frame_num
        angle_sound.sound_id1 = sound_id
        angle_sound.wwise_sound_id1 = wwise_sound_id
        angle_sound.sound_delay2 = frame_num_2
        angle_sound.sound_id2 = sound_id_2
        angle_sound.wwise_sound_id2 = wwise_sound_id_2
        angle_sound.sound_delay3 = frame_num_3
        angle_sound.sound_id3 = sound_id_3
        angle_sound.wwise_sound_id3 = wwise_sound_id_3
        
        # CRITICAL: bfp_rs validation is strict on list assignment vs num_facets.
        # Strategy: Use Reset-Append to bypass assignment check and break storage sharing.
        
        # 1. Backup existing sounds (triggers copy)
        current_sounds = list(self._sprite.facet_attack_sounds)
        current_sounds.append(angle_sound)
        
        # 2. Clear to break storage linkage & satisfy "Expected: 0" check if uninitialized
        self._sprite.facet_attack_sounds = []
        
        # 3. Append individually (bypasses strict size check)
        for sound in current_sounds:
            self._sprite.facet_attack_sounds.append(sound)
            
        # 4. Sync metadata
        self._sprite.facets_have_attack_sounds = True
        self._sprite.num_facets = len(self._sprite.facet_attack_sounds)
    
    @property
    def deltas(self) -> list[DeltaHandle]:
        """
        Get all deltas as DeltaHandle objects.
        
        Returns:
            List of DeltaHandle objects for iteration
        """
        from aoe2_genie_tooling.Graphics.delta_handle import DeltaHandle
        return [DeltaHandle(self, i) for i in range(len(self._sprite.deltas))]
    
    def remove_delta_by_graphic(self, graphic_id: int) -> int:
        """
        Remove all deltas from THIS graphic that point to a specific graphic ID.
        
        Args:
            graphic_id: Graphic ID to search for
            
        Returns:
            Number of deltas removed
        """
        initial_count = len(self._sprite.deltas)
        self._sprite.deltas = [
            d for d in self._sprite.deltas if d.sprite_id != graphic_id
        ]
        removed = initial_count - len(self._sprite.deltas)
        if removed > 0:
            self._sprite.num_deltas = len(self._sprite.deltas)
        return removed
    
    def clear_angle_sounds(self) -> None:
        """Remove all angle sounds."""
        self._sprite.facet_attack_sounds.clear()
        self._sprite.facets_have_attack_sounds = False



"""
Orchestration logic for RECodeGenerator.
"""
from __future__ import annotations
from pathlib import Path
from typing import Any, Dict, List
from .models import LinkSpec, IndependentObjects
from .utils import safe_name


def create_folder_structure(folder_location: Path, groups: List[Any], independent: IndependentObjects) -> None:
    """Create the folder structure with groups/ and independent/ subdirectories."""
    folder_location.mkdir(parents=True, exist_ok=True)
    
    # Create groups/ subdirectory
    groups_folder = folder_location / "groups"
    groups_folder.mkdir(exist_ok=True)
    
    # Create independent/ subdirectory
    independent_folder = folder_location / "independent"
    independent_folder.mkdir(exist_ok=True)
    
    # Generate main __init__.py
    init_lines = [
        '"""',
        'Generated by RECodeGenerator',
        '',
        f'Contains {len(groups)} groups with units, graphics, and sounds.',
        '"""',
        'import sys',
        'from pathlib import Path',
        '',
        '# Add this directory to path for imports',
        '_pkg_dir = Path(__file__).parent',
        'if str(_pkg_dir) not in sys.path:',
        '    sys.path.insert(0, str(_pkg_dir))',
        '',
        '# Import all apply functions for easy access',
    ]
    
    for group in groups:
        safe_grp_name = safe_name(group.name)
        init_lines.append(
            f'from groups.{group.folder_name}.apply_{safe_grp_name} import apply_{safe_grp_name}_group'
        )
    
    init_lines.extend([
        '',
        '# Import master module',
        'from reverse_engineer_dat import reverse_engineer_dat, list_groups',
        '',
    ])
    
    (folder_location / "__init__.py").write_text("\n".join(init_lines))
    
    # Create groups/__init__.py
    groups_init_lines = [
        '"""Groups containing related units, graphics, and sounds."""',
    ]
    (groups_folder / "__init__.py").write_text("\n".join(groups_init_lines))
    
    # Create group folders inside groups/
    for group in groups:
        group_folder = groups_folder / group.folder_name
        group_folder.mkdir(exist_ok=True)
        
        safe_grp_name = safe_name(group.name)
        group_init_lines = [
            f'"""Group: {group.name}"""',
            'import sys',
            'from pathlib import Path',
            '',
            '_group_dir = Path(__file__).parent',
            'if str(_group_dir) not in sys.path:',
            '    sys.path.insert(0, str(_group_dir))',
            '',
            f'from apply_{safe_grp_name} import apply_{safe_grp_name}_group',
            'from units import *',
            'from graphics import *', 
            'from sounds import *',
            '',
        ]
        
        (group_folder / "__init__.py").write_text("\n".join(group_init_lines))
    
    # Create independent/__init__.py
    independent_init_lines = [
        '"""Independent objects with no links to other above-threshold objects."""',
        'import sys',
        'from pathlib import Path',
        '',
        '_indep_dir = Path(__file__).parent',
        'if str(_indep_dir) not in sys.path:',
        '    sys.path.insert(0, str(_indep_dir))',
        '',
        'from units import *',
        'from graphics import *',
        'from sounds import *',
        '',
    ]
    (independent_folder / "__init__.py").write_text("\n".join(independent_init_lines))


def generate_apply_module(ws: Any, group: Any, folder_location: Path,
                         all_unit_links: Dict[int, List[LinkSpec]],
                         all_graphic_links: Dict[int, List[LinkSpec]],
                         all_sound_links: Dict[int, List[LinkSpec]]) -> None:
    """Generate the apply_group_X.py module."""
    group_folder = folder_location / "groups" / group.folder_name
    
    lines = [
        '"""',
        f'Apply Group: {group.name}',
        f'',
        f'Creates all objects and links them together.',
        '"""',
        'import sys',
        'import importlib.util',
        'from pathlib import Path',
        '',
        '# Load local modules',
        '_current_dir = Path(__file__).parent',
        '',
        'def _load_local_module(name):',
        '    spec = importlib.util.spec_from_file_location(name, _current_dir / f"{name}.py")',
        '    mod = importlib.util.module_from_spec(spec)',
        '    spec.loader.exec_module(mod)',
        '    return mod',
        '',
        '_units = _load_local_module("units")',
        '_graphics = _load_local_module("graphics")',
        '_sounds = _load_local_module("sounds")',
        '',
        '',
        f'def apply_{safe_name(group.name)}_group(ws):',
        f'    """Create and link all objects in {group.name} group."""',
        f'    print(f"  > Applying Group: {group.name}")',
        '',
    ]
    
    # 1. Create Graphics
    lines.append("    # Create Graphics")
    graphic_vars = {}
    for gid in sorted(group.graphic_ids):
        try:
            graphic = ws.graphic_manager.get(gid)
            name = getattr(graphic, "name", f"Graphic_{gid}")
        except Exception:
            name = f"Graphic_{gid}"
        
        gfx_safe_name = safe_name(name)
        var_name = f"graphic_{gid}"
        graphic_vars[gid] = var_name
        lines.append(f"    {var_name} = _graphics.create_graphic_{gid}_{gfx_safe_name}(ws)")
        
    lines.append("")
    
    # 2. Create Sounds
    lines.append("    # Create Sounds")
    sound_vars = {}
    for sid in sorted(group.sound_ids):
        try:
            sound = ws.sound_manager.get(sid)
            sound_name = f"sound_{sid}"
            sound_files = sound.sounds if hasattr(sound, 'sounds') else []
            if sound_files:
                first_file = sound_files[0]
                if hasattr(first_file, 'sound_name') and first_file.sound_name:
                    sound_name = first_file.sound_name
                elif hasattr(first_file, 'filename') and first_file.filename:
                    sound_name = first_file.filename
        except Exception:
            sound_name = f"sound_{sid}"

        snd_safe_name = safe_name(sound_name)
        var_name = f"sound_{sid}"
        sound_vars[sid] = var_name
        lines.append(f"    {var_name} = _sounds.create_sound_{sid}_{snd_safe_name}(ws)")
        
    lines.append("")
    
    # 3. Create Units
    lines.append("    # Create Units")
    unit_vars = {}
    for uid in sorted(group.unit_ids):
        try:
            unit = ws.unit_manager.get(uid)
            name = getattr(unit, "name", f"Unit_{uid}")
        except Exception:
            name = f"Unit_{uid}"
        
        unit_safe_name = safe_name(name)
        var_name = f"unit_{uid}"
        unit_vars[uid] = var_name
        lines.append(f"    {var_name} = _units.create_unit_{uid}_{unit_safe_name}(ws)")
        
    lines.append("")
    lines.append("    # Apply Links")
    
    # Link Graphics
    for gid in sorted(group.graphic_ids):
        links = all_graphic_links.get(gid, [])
        if not links:
            continue

        var_name = graphic_vars.get(gid, f"graphic_{gid}")
        for link in sorted(links, key=lambda l: l.field):
            setter = f"{var_name}.{link.field}"

            if link.target_type == "graphic" and link.target_id in graphic_vars:
                lines.append(f"    {setter} = {graphic_vars[link.target_id]}.id")
            elif link.target_type == "sound" and link.target_id in sound_vars:
                lines.append(f"    {setter} = {sound_vars[link.target_id]}.id")
            else:
                lines.append(f"    {setter} = {link.target_id}  # external {link.target_type}")

    # Link Sounds
    for sid in sorted(group.sound_ids):
        links = all_sound_links.get(sid, [])
        if not links:
            continue

        var_name = sound_vars.get(sid, f"sound_{sid}")
        for link in sorted(links, key=lambda l: l.field):
            setter = f"{var_name}.{link.field}"
            lines.append(f"    {setter} = {link.target_id}")

    # Link Units
    for uid in sorted(group.unit_ids):
        links = all_unit_links.get(uid, [])
        if not links:
            continue

        var_name = unit_vars.get(uid, f"unit_{uid}")

        for link in sorted(links, key=lambda l: l.field):
            setter = f"{var_name}.{link.field}"

            if link.target_type == "unit" and link.target_id in unit_vars:
                lines.append(f"    {setter} = {unit_vars[link.target_id]}.id")
            elif link.target_type == "graphic" and link.target_id in graphic_vars:
                lines.append(f"    {setter} = {graphic_vars[link.target_id]}.id")
            elif link.target_type == "sound" and link.target_id in sound_vars:
                lines.append(f"    {setter} = {sound_vars[link.target_id]}.id")
            else:
                lines.append(f"    {setter} = {link.target_id}  # external {link.target_type}")
    
    lines.append("")
    lines.append("    # Return main unit")
    if unit_vars:
        first_var = list(unit_vars.values())[0]
        lines.append(f"    return {first_var}")
    else:
        lines.append("    return None")
    lines.append("")
    
    apply_path = group_folder / f"apply_{safe_name(group.name)}.py"
    apply_path.write_text("\n".join(lines))


def generate_master_module(folder_location: Path, groups: List[Any], independent: IndependentObjects) -> None:
    """Generate the master reverse_engineer_dat.py module."""
    lines = [
        '"""',
        'Reverse Engineer DAT - Master Apply Module',
        '',
        f'Generated {len(groups)} groups',
        'Master Module: reverse_engineer_dat',
        '',
        'Orchestrates the reverse engineering of all grouped units.',
        '"""',
        'import sys',
        'import os',
        'from pathlib import Path',
        '',
        '# Add current directory to sys.path',
        'base_dir = Path(__file__).resolve().parent',
        'if str(base_dir) not in sys.path:',
        '    sys.path.insert(0, str(base_dir))',
        '',
        'from aoe2_genie_tooling.Base.workspace import GenieWorkspace',
        '',
    ]
    
    # Import group apply functions
    for group in groups:
        safe_grp_name = safe_name(group.name)
        lines.extend([
            f'try:',
            f'    from groups.{group.folder_name}.apply_{safe_grp_name} import apply_{safe_grp_name}_group',
            f'except ImportError:',
            f'    from .groups.{group.folder_name}.apply_{safe_grp_name} import apply_{safe_grp_name}_group'
        ])
    
    # Import independent objects
    lines.extend([
        '',
        '# Import independent objects',
        'try:',
        '    from independent.units import *',
        '    from independent.graphics import *',
        '    from independent.sounds import *',
        'except ImportError:',
        '    from .independent.units import *',
        '    from .independent.graphics import *',
        '    from .independent.sounds import *',
    ])
    
    lines.extend([
        '',
        '',
        'def reverse_engineer_dat(ws):',
        '    """Apply all groups and independent objects to recreate units."""',
        '    results = {}',
        f'    total_groups = {len(groups)}',
        '    ',
    ])
    
    # Apply independent objects first
    lines.append('    # Apply independent objects')
    lines.append('    print("Applying independent objects...")')
    # Note: We'll add specific calls when generating the master - this is done at generate time
    
    lines.append('')
    lines.append('    # Apply groups')
    for i, group in enumerate(groups):
        safe_grp_name = safe_name(group.name)
        lines.append(f'    print(f"[{i+1}/{{total_groups}}] Processing group: {group.name}...")')
        lines.append(f'    results["{safe_grp_name}"] = apply_{safe_grp_name}_group(ws)')
    
    lines.extend([
        '',
        '    return results',
        '',
        'def list_groups():',
        '    """List names of all generated groups."""',
        '    return [',
    ])
    
    for group in groups:
        lines.append(f'        "{group.name}",')
        
    lines.extend([
        '    ]',
        '',
        '',
        'if __name__ == "__main__":',
        '    import argparse',
        '    ',
        '    parser = argparse.ArgumentParser(description="Reverse engineer DAT file")',
        '    parser.add_argument("--input", "-i", default="empires2_x2_p1.dat", help="Input DAT file")',
        '    parser.add_argument("--output", "-o", default="Rebuild_Output.dat", help="Output DAT file")',
        '    args = parser.parse_args()',
        '    ',
        '    print(f"Loading workspace from {args.input}...")',
        '    ws = GenieWorkspace.load(args.input)',
        '    ',
        '    print(f"Applying {len(list_groups())} groups...")',
        '    results = reverse_engineer_dat(ws)',
        '    ',
        '    print(f"Saving to {args.output}...")',
        '    ws.save(args.output, validate=False)',
        '    print("Done!")',
    ])
    
    (folder_location / "reverse_engineer_dat.py").write_text("\n".join(lines))


def generate_group_code(ws: Any, group: Any, folder_location: Path, config: Any) -> None:
    """Generate all code for a group."""
    from .writers import generate_unit_code, generate_graphic_code, generate_sound_code
    
    group_folder = folder_location / "groups" / group.folder_name
    
    # Track all links
    all_unit_links: Dict[int, List[LinkSpec]] = {}
    all_graphic_links: Dict[int, List[LinkSpec]] = {}
    all_sound_links: Dict[int, List[LinkSpec]] = {}
    
    # ===== UNITS =====
    units_lines = [
        '"""',
        f'Units for group: {group.name}',
        f'Unit IDs: {sorted(group.unit_ids)}',
        '"""',
        'from aoe2_genie_tooling.Base.workspace import GenieWorkspace',
        '',
    ]
    
    for unit_id in sorted(group.unit_ids):
        unit_code, links = generate_unit_code(ws, unit_id, group, config)
        units_lines.extend(unit_code)
        all_unit_links[unit_id] = links
    
    (group_folder / "units.py").write_text("\n".join(units_lines))
    
    # ===== GRAPHICS =====
    graphics_lines = [
        '"""',
        f'Graphics for group: {group.name}',
        f'Graphic IDs: {sorted(group.graphic_ids)}',
        '"""',
        'from aoe2_genie_tooling.Base.workspace import GenieWorkspace',
        '',
    ]
    
    for gid in sorted(group.graphic_ids):
        gfx_code, links = generate_graphic_code(ws, gid, group, config)
        graphics_lines.extend(gfx_code)
        all_graphic_links[gid] = links
    
    (group_folder / "graphics.py").write_text("\n".join(graphics_lines))
    
    # ===== SOUNDS =====
    sounds_lines = [
        '"""',
        f'Sounds for group: {group.name}',
        f'Sound IDs: {sorted(group.sound_ids)}',
        '"""',
        'from aoe2_genie_tooling.Base.workspace import GenieWorkspace',
        '',
    ]
    
    for sid in sorted(group.sound_ids):
        snd_code, links = generate_sound_code(ws, sid, group, config)
        sounds_lines.extend(snd_code)
        all_sound_links[sid] = links
    
    (group_folder / "sounds.py").write_text("\n".join(sounds_lines))
    
    # ===== APPLY MODULE =====
    generate_apply_module(ws, group, folder_location, all_unit_links, all_graphic_links, all_sound_links)


def generate_independent_code(ws: Any, independent: IndependentObjects, folder_location: Path, config: Any) -> None:
    """Generate code for independent objects."""
    from .writers import (
        generate_independent_unit_code,
        generate_independent_graphic_code,
        generate_independent_sound_code,
    )
    
    independent_folder = folder_location / "independent"
    
    # ===== UNITS =====
    units_lines = [
        '"""',
        'Independent Units',
        f'Unit IDs: {sorted(independent.unit_ids)}',
        '"""',
        'from aoe2_genie_tooling.Base.workspace import GenieWorkspace',
        '',
    ]
    
    for unit_id in sorted(independent.unit_ids):
        unit_code = generate_independent_unit_code(ws, unit_id, config)
        units_lines.extend(unit_code)
    
    (independent_folder / "units.py").write_text("\n".join(units_lines))
    
    # ===== GRAPHICS =====
    graphics_lines = [
        '"""',
        'Independent Graphics',
        f'Graphic IDs: {sorted(independent.graphic_ids)}',
        '"""',
        'from aoe2_genie_tooling.Base.workspace import GenieWorkspace',
        '',
    ]
    
    for gid in sorted(independent.graphic_ids):
        gfx_code = generate_independent_graphic_code(ws, gid, config)
        graphics_lines.extend(gfx_code)
    
    (independent_folder / "graphics.py").write_text("\n".join(graphics_lines))
    
    # ===== SOUNDS =====
    sounds_lines = [
        '"""',
        'Independent Sounds',
        f'Sound IDs: {sorted(independent.sound_ids)}',
        '"""',
        'from aoe2_genie_tooling.Base.workspace import GenieWorkspace',
        '',
    ]
    
    for sid in sorted(independent.sound_ids):
        snd_code = generate_independent_sound_code(ws, sid, config)
        sounds_lines.extend(snd_code)
    
    (independent_folder / "sounds.py").write_text("\n".join(sounds_lines))



"""
Base module containing GenieWorkspace - the root entry point.

This module provides the main workspace class for loading, editing,
and saving Genie Engine DAT files.
"""

from aoe2_genie_tooling.Base.workspace import GenieWorkspace

__all__ = ["GenieWorkspace"]



"""
CivHandle - Wrapper for individual Civilization objects.

Civilizations contain units, resources, and effect links.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    from aoe2_genie_tooling.Base.workspace import GenieWorkspace
    from aoe2_genie_tooling.Civilizations.resource_accessor import ResourceAccessor

__all__ = ["CivHandle"]


class CivHandle:
    """
    Handle for a single civilization.
    
    Provides direct attribute access and resource management.
    """
    
    def __init__(self, workspace: GenieWorkspace, civ_id: int) -> None:
        """Initialize CivHandle."""
        object.__setattr__(self, '_workspace', workspace)
        object.__setattr__(self, '_id', civ_id)
        object.__setattr__(self, '_civ', workspace.dat.civilizations[civ_id])
    
    @property
    def id(self) -> int:
        """Get the civilization ID."""
        return self._id
    
    @property
    def workspace(self) -> GenieWorkspace:
        """Get the workspace."""
        return self._workspace

    @property
    def name(self) -> str:
        """Get the civilization name."""
        try:
            return self._civ.name
        except Exception:
            return ""

    @name.setter
    def name(self, value: str) -> None:
        """Set the civilization name."""
        try:
            self._civ.name = value
        except Exception:
            pass

    @property
    def tech_tree_effect_id(self) -> int:
        """Get the tech tree effect ID."""
        return self._civ.tech_tree_effect_id

    @tech_tree_effect_id.setter
    def tech_tree_effect_id(self, value: int) -> None:
        self._civ.tech_tree_effect_id = value

    @property
    def team_bonus_effect_id(self) -> int:
        """Get the team bonus effect ID."""
        try:
            return self._civ.team_bonus_effect_id
        except Exception:
            return -1

    @team_bonus_effect_id.setter
    def team_bonus_effect_id(self, value: int) -> None:
        try:
            self._civ.team_bonus_effect_id = value
        except Exception:
            pass

    @property
    def icon_set(self) -> int:
        """Get the icon set."""
        return self._civ.icon_set

    @icon_set.setter
    def icon_set(self, value: int) -> None:
        self._civ.icon_set = value

    @property
    def player_type(self) -> int:
        """Get the player type."""
        return self._civ.player_type

    @player_type.setter
    def player_type(self, value: int) -> None:
        self._civ.player_type = value

    # Resource Management
    @property
    def resources(self) -> list:
        """Get the resources list directly."""
        return self._civ.resources

    @property
    def resource(self) -> ResourceAccessor:
        """
        Access per-civ resource values.
        
        Usage:
            civ.resource.get(0)           # Get value at index 0
            civ.resource.set(0, 200.0)    # Set value at index 0
            civ.resource[0]               # Get via indexing
            civ.resource[0] = 200.0       # Set via indexing
            
        Note: To add/remove resources, use CivManager.add_resource() and
              CivManager.remove_resource() - these apply to ALL civilizations.
        """
        from aoe2_genie_tooling.Civilizations.resource_accessor import ResourceAccessor
        return ResourceAccessor(self)

    # Unit Management
    @property
    def units(self) -> list:
        """Get the units list."""
        return self._civ.units

    def exists(self) -> bool:
        """Check if this civilization entry exists."""
        return self._civ is not None

    def __getattr__(self, name: str) -> Any:
        """Get attribute from underlying civilization."""
        return getattr(self._civ, name)
    
    def __setattr__(self, name: str, value: Any) -> None:
        """Set attribute on underlying civilization."""
        if name.startswith('_'):
            object.__setattr__(self, name, value)
        else:
            setattr(self._civ, name, value)
    
    def __repr__(self) -> str:
        if not self.exists():
            return f"CivHandle(id={self._id}, status=DELETED)"
        return f"CivHandle(id={self._id}, name='{self.name}')"



from __future__ import annotations

from bfp_rs import BaseStruct, Retriever, Version
from bfp_rs.types.le import Array, i16, f32, i8, i32, u8, Array16

from sections.civilization.type_info.unit_cost import UnitCost

class TrainLocation(BaseStruct):
    # @formatter:off
    train_time: int                             = Retriever(i16,                                         default = 0)
    location_unit_id: int                       = Retriever(i16,                                         default = -1)
    button_id: int                              = Retriever(u8,                                          default = 0)
    hotkey_id: int                              = Retriever(i32,                min_ver = Version(8, 7), default = 0)
    # @formatter:on

class CreationInfo(BaseStruct):
    # @formatter:off
    costs: list[UnitCost]                       = Retriever(Array[3][UnitCost],                                                         default_factory = lambda ver: [UnitCost(ver) for _ in range(3)])

    train_locations_old: list[TrainLocation]    = Retriever(Array[1][TrainLocation],                          max_ver = Version(8, 5),  default_factory = lambda ver: [TrainLocation(ver)])
    train_locations_new: list[TrainLocation]    = Retriever(Array16[TrainLocation],  min_ver = Version(8, 6),                           default_factory = lambda ver: [])

    rear_attack_modifier: float                 = Retriever(f32,                     min_ver = Version(5, 7), default = 0)
    """aka heal_timer"""
    flank_attack_modifier: float                = Retriever(f32,                     min_ver = Version(5, 7), default = 0)
    creatable_type: int                         = Retriever(i8,                      min_ver = Version(5, 7), default = 0)
    hero_mode: int                              = Retriever(i8,                      min_ver = Version(5, 7), default = 0)

    garrisoned_sprite_id: int                   = Retriever(i32,                     min_ver = Version(5, 7), default = -1)
    spawning_sprite_id: int                     = Retriever(i16,                     min_ver = Version(7, 1), default = -1)
    upgrading_sprite_id: int                    = Retriever(i16,                     min_ver = Version(7, 1), default = -1)
    hero_glowing_sprite_id: int                 = Retriever(i16,                     min_ver = Version(7, 4), default = -1)
    idle_attack_graphic: int                    = Retriever(i16,                     min_ver = Version(8, 3), default = -1)
    max_charge: float                           = Retriever(f32,                     min_ver = Version(7, 3), default = 0)
    charge_regen_rate: float                    = Retriever(f32,                     min_ver = Version(7, 3), default = 0)

    charge_event: int                           = Retriever(i16,                     min_ver = Version(7, 3), default = 0)
    charge_type: int                            = Retriever(i16,                     min_ver = Version(7, 3), default = 0)
    charge_target: float                        = Retriever(i16,                     min_ver = Version(8, 0), default = 0)
    charge_projectile_unit: float               = Retriever(i32,                     min_ver = Version(8, 2), default = 0)
    attack_priority: float                      = Retriever(u8,                      min_ver = Version(8, 2), default = 0)
    invulnerability_level: float                = Retriever(f32,                     min_ver = Version(8, 2), default = 0)

    button_icon_id: float                       = Retriever(i16,                     min_ver = Version(8, 1), default = 0)
    button_short_tooltip_str_id: float          = Retriever(i32,                     min_ver = Version(8, 1), default = 0)
    button_extend_tooltip_str_id: float         = Retriever(i32,                     min_ver = Version(8, 1), default = 0)
    button_hotkey_action: float                 = Retriever(i16,                     min_ver = Version(8, 1), default = 0)


    min_conversion_time_modifier: float         = Retriever(f32,                     min_ver = Version(7, 6), default = 0)
    max_conversion_time_modifier: float         = Retriever(f32,                     min_ver = Version(7, 6), default = 0)
    conversion_chance_mod: float                = Retriever(f32,                     min_ver = Version(7, 6), default = 0)

    min_projectiles: float                      = Retriever(f32,                     min_ver = Version(5, 7), default = 0)
    max_projectiles: int                        = Retriever(i8,                      min_ver = Version(5, 7), default = 0)
    projectile_spawning_area_width: float       = Retriever(f32,                     min_ver = Version(5, 7), default = 0)
    projectile_spawning_area_length: float      = Retriever(f32,                     min_ver = Version(5, 7), default = 0)
    projectile_spawning_area_randomness: float  = Retriever(f32,                     min_ver = Version(5, 7), default = 0)
    secondary_projectile_unit_id: int           = Retriever(i32,                     min_ver = Version(5, 7), default = 0)
    special_graphic_id: int                     = Retriever(i32,                     min_ver = Version(5, 7), default = 0)
    special_activation: int                     = Retriever(i8,                      min_ver = Version(5, 7), default = 0)

    displayed_pierce_armor: int                 = Retriever(i16,                                              default = 0)
    # @formatter:on



from .animated_info import AnimationInfo
from .building_annex import BuildingAnnex
from .building_info import BuildingInfo
from .combat_info import CombatInfo
from .creation_info import CreationInfo
from .damage_class import DamageClass
from .looting_table import LootingTable
from .movement_info import MovementInfo
from .projectile_info import ProjectileInfo
from .task_info import TaskInfo
from .unit_cost import UnitCost
from .unit_type import UnitType



"""Type stubs for CivManager - enables IDE autocomplete"""
from typing import Optional, Any
from aoe2_genie_tooling.Civilizations.civ_handle import CivHandle

class CivManager:
    """Manager for civilization operations."""
    
    def get(self, civ_id: int) -> CivHandle:
        """Get a civilization by ID."""
        ...
    
    def count(self) -> int:
        """Get total number of civilizations."""
        ...
    
    def exists(self, civ_id: int) -> bool:
        """Check if civilization exists."""
        ...

    def find_by_name(self, name: str) -> Optional[CivHandle]:
        """Find first civilization matching name."""
        ...

    def add_new(
        self,
        name: str = "",
        civ_id: Optional[int] = None,
    ) -> CivHandle:
        """Add a new civilization."""
        ...

    def create(
        self,
        name: str = "",
        civ_id: Optional[int] = None,
    ) -> CivHandle:
        """Alias for add_new."""
        ...

    def copy(self, source_id: int, target_id: Optional[int] = None) -> CivHandle:
        """Copy a civilization to a new ID."""
        ...

    def copy_to_clipboard(self, civ_id: int) -> bool:
        """Copy civilization to internal clipboard."""
        ...

    def paste(self, target_id: Optional[int] = None) -> Optional[CivHandle]:
        """Paste civilization from clipboard."""
        ...

    def clear_clipboard(self) -> None:
        """Clear clipboard."""
        ...

    # Global Resource Management
    def add_resource(self, default_value: float = 0.0) -> int:
        """
        Add a new resource to ALL civilizations.
        
        Resources are global - this adds a slot to every civilization.
        """
        ...

    def remove_resource(self, index: int) -> bool:
        """
        Remove a resource from ALL civilizations.
        
        Resources are global - this removes from every civilization.
        """
        ...

    def resource_count(self) -> int:
        """Get the number of resources (same across all civs)."""
        ...

    def clear_resources(self) -> None:
        """Remove all resources from ALL civilizations."""
        ...



"""
Terrain types dataset.

Contains terrain IDs used for foundation_terrain, placement_terrain, etc.
"""
from enum import IntEnum

__all__ = ["TerrainType"]


class TerrainType(IntEnum):
    """
    Terrain type IDs for foundation_terrain, placement_terrain, terrain_restriction.
    
    Note: Full terrain list can have 100+ terrains depending on the dataset.
    These are common base game terrains.
    """
    GRASS = 0
    WATER_SHALLOW = 1
    BEACH = 2
    DIRT = 3
    WATER_MEDIUM = 4
    LEAVES = 5
    DIRT_3 = 6
    FARM = 7
    DEAD_FARM = 8
    GRASS_2 = 9
    FOREST_LEAVES = 10
    DIRT_4 = 11
    GRASS_3 = 12
    FOREST_SNOW = 13
    DESERT = 14
    WATER_DEEP = 15
    GRASS_4 = 16
    JUNGLE_LEAVES = 17
    BAMBOO = 18
    GRASS_5 = 19
    BAMBOO_2 = 20
    PALM_DESERT = 21
    WATER_SHALLOW_2 = 22
    MANGROVE_FOREST = 23
    ROAD = 24
    ROAD_2 = 25
    WATER_BRIDGE = 26
    FARM_CONSTRUCT = 27
    FARM_2 = 28
    FOUNDATION = 29  # Building foundation terrain
    GRASS_6 = 30
    WATER_SHALLOW_3 = 31
    SNOW = 32
    SNOW_FOREST = 33
    SNOW_2 = 34
    SNOW_3 = 35
    ICE = 36
    SNOW_ROAD = 37
    FUNGUS_ROAD = 38
    KOH = 39  # King of the Hill
    WATER_FORD = 40
    FARM_DEAD = 41
    # ... additional terrains 42+ in newer datasets



from __future__ import annotations

from bfp_rs import BaseStruct, Retriever, Version, RetrieverCombiner
from bfp_rs.types.le import Bytes, str16, NtStr, Array16

from sections.tech_effect.effect_command import EffectCommand


class TechEffect(BaseStruct):
    # @formatter:off
    _str_sign_de1: bytes         = Retriever(Bytes[2],  min_ver = Version(4, 5), max_ver = Version(4, 5), default = b"\x0A\x60")
    _name_de1: str               = Retriever(str16,     min_ver = Version(4, 5), max_ver = Version(4, 5), default = "")

    _str_sign_de2: bytes         = Retriever(Bytes[2],  min_ver = Version(7, 1),                          default = b"\x0A\x60")
    _name_de2: str               = Retriever(str16,     min_ver = Version(7, 1),                          default = "")

    _name_aoe1: str              = Retriever(NtStr[31], min_ver = Version(3, 7), max_ver = Version(3, 7), default = "")
    _name_aoe2_swgb: str         = Retriever(NtStr[31], min_ver = Version(5, 7), max_ver = Version(5, 9), default = "")

    name: str                    = RetrieverCombiner(_name_de2, _name_aoe2_swgb, _name_de1, _name_aoe1)

    effects: list[EffectCommand] = Retriever(Array16[EffectCommand],                                                   default_factory = lambda _: [])
    # @formatter:on



from __future__ import annotations
from typing import TYPE_CHECKING, List

from ..Shared.tool_base import ToolBase
from ..backend import DatFile

if TYPE_CHECKING:
    from ..backend import Terrain

class TerrainManager(ToolBase):
    """
    Manager for accessing terrain data in a DAT file.
    """

    def __init__(self, dat_file: DatFile) -> None:
        super().__init__(dat_file)
        self._terrains = self.dat_file.terrains

    def get_terrains(self) -> List[Terrain]:
        """
        Get a list of all terrains.

        Returns:
            A list of terrain objects.
        """
        return self._terrains

    def get_terrain(self, terrain_id: int) -> Terrain:
        """
        Get a single terrain by its ID.

        Args:
            terrain_id: The ID of the terrain to retrieve.

        Returns:
            The terrain object.

        Raises:
            IndexError: If the terrain_id is out of bounds.
        """
        return self._terrains[terrain_id]

    def count(self) -> int:
        """
        Return the total number of terrains.
        """
        return len(self._terrains)



from __future__ import annotations

from bfp_rs import BaseStruct, Retriever
from bfp_rs.types.le import i16, i8


class TechCost(BaseStruct):
    # @formatter:off
    resource_id: int  = Retriever(i16,    default = 0)
    quantity: int     = Retriever(i16,    default = 0)
    deduct_flag: bool = Retriever(i8,     default = 0)
    # @formatter:on



"""
SoundFileHandle - Wrapper for individual SoundFile objects.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from aoe2_genie_tooling.Sounds.sound_handle import SoundHandle

__all__ = ["SoundFileHandle"]


class SoundFileHandle:
    """
    Handle for a single sound file entry within a sound.
    """
    
    def __init__(self, parent_handle: SoundHandle, file_id: int) -> None:
        """
        Initialize SoundFileHandle.
        
        Args:
            parent_handle: The SoundHandle owning this file
            file_id: Index in the sound_files list
        """
        object.__setattr__(self, '_parent', parent_handle)
        object.__setattr__(self, '_id', file_id)
        # Access the raw Sound object from parent and get the SoundFile
        object.__setattr__(self, '_file', parent_handle._sound.sound_files[file_id])
    
    @property
    def index(self) -> int:
        """Get the index of this file in the sound's list."""
        return self._id
    
    @property
    def filename(self) -> str:
        """Get filename with version fallback."""
        try:
            return self._file.filename
        except Exception:
            try:
                return self._file.sound_name
            except Exception:
                return ""

    @filename.setter
    def filename(self, value: str) -> None:
        """Set filename with version fallback."""
        try:
            self._file.filename = value
        except Exception:
            try:
                self._file.sound_name = value
            except Exception:
                pass

    @property
    def sound_name(self) -> str:
        """Get sound_name with version fallback."""
        try:
            return self._file.sound_name
        except Exception:
            try:
                return self._file.filename
            except Exception:
                return ""

    @sound_name.setter
    def sound_name(self, value: str) -> None:
        """Set sound_name with version fallback."""
        try:
            self._file.sound_name = value
        except Exception:
            try:
                self._file.filename = value
            except Exception:
                pass

    def __getattr__(self, name: str) -> Any:
        return getattr(self._file, name)
    
    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith('_'):
            object.__setattr__(self, name, value)
        else:
            setattr(self._file, name, value)

    def __repr__(self) -> str:
        return f"SoundFileHandle(index={self._id}, filename='{getattr(self, 'filename', 'Unknown')}')"



"""Type stubs for TechHandle - enables IDE autocomplete"""
from typing import Any, Optional, Tuple


class CostBuilder:
    """Builder for setting tech costs."""
    
    def cost_1(self, resource_type: int, amount: int, deduct: bool = True) -> None:
        """
        Set cost slot 1.
        
        Args:
            resource_type: 0=Food, 1=Wood, 2=Stone, 3=Gold
            amount: Quantity
            deduct: Whether to deduct (default True)
        """
        ...
    
    def cost_2(self, resource_type: int, amount: int, deduct: bool = True) -> None:
        """Set cost slot 2."""
        ...
    
    def cost_3(self, resource_type: int, amount: int, deduct: bool = True) -> None:
        """Set cost slot 3."""
        ...


class TechHandle:
    """Handle for a single tech."""
    
    @property
    def id(self) -> int:
        """Get the tech ID."""
        ...
    
    @property
    def set_cost(self) -> CostBuilder:
        """
        Builder for setting costs.
        
        Example:
            tech.set_cost.cost_1(3, 100)  # 100 Gold
            tech.set_cost.cost_2(0, 50)   # 50 Food
        """
        ...

    @property
    def name(self) -> str:
        """Get the tech name."""
        ...

    @name.setter
    def name(self, value: str) -> None:
        ...

    @property
    def effect_id(self) -> int:
        """Get the linked effect ID."""
        ...

    @effect_id.setter
    def effect_id(self, value: int) -> None:
        ...

    @property
    def research_time(self) -> int:
        """Get research time."""
        ...

    @research_time.setter
    def research_time(self, value: int) -> None:
        ...

    @property
    def cost_1(self) -> Any:
        """Get first cost slot."""
        ...

    @property
    def cost_2(self) -> Any:
        """Get second cost slot."""
        ...

    @property
    def cost_3(self) -> Any:
        """Get third cost slot."""
        ...

    @property
    def costs(self) -> list:
        """Get the cost array (3 TechCost objects). Use set_cost.cost_1() etc for easier modification."""
        ...

    def clear_cost(self, slot: int) -> None:
        """Clear a cost slot (set to 0 quantity)."""
        ...

    def clear_all_costs(self) -> None:
        """Clear all cost slots."""
        ...

    def get_cost(self, slot: int) -> Tuple[int, int, int]:
        """Get cost at slot as tuple (resource_id, quantity, deduct_flag)."""
        ...

    @property
    def required_tech_ids(self) -> tuple:
        """Get required tech IDs (6 slots, -1 means empty). Use set_required_tech() to modify."""
        ...

    def set_required_tech(self, slot: int, tech_id: int) -> None:
        """
        Set a required tech ID at a specific slot.
        
        Args:
            slot: Slot index (0-5)
            tech_id: Tech ID to require (-1 to clear the slot)
            
        Raises:
            ValueError: If slot is not in range 0-5
        """
        ...

    def clear_required_techs(self) -> None:
        """Clear all required tech slots (set to -1)."""
        ...

    @property
    def icon_id(self) -> int:
        """Get icon ID."""
        ...

    @icon_id.setter
    def icon_id(self, value: int) -> None:
        ...

    @property
    def type(self) -> int:
        """Get tech type."""
        ...

    @type.setter
    def type(self, value: int) -> None:
        ...

    @property
    def min_required_techs(self) -> int:
        """Get minimum required techs count."""
        ...

    @min_required_techs.setter
    def min_required_techs(self, value: int) -> None:
        ...

    @property
    def civilization_id(self) -> int:
        """Get civilization ID (-1 = all civs)."""
        ...

    @civilization_id.setter
    def civilization_id(self, value: int) -> None:
        ...

    @property
    def full_tech_tree_mode(self) -> int:
        """Get full tech tree mode."""
        ...

    @full_tech_tree_mode.setter
    def full_tech_tree_mode(self, value: int) -> None:
        ...

    @property
    def location_unit_id(self) -> int:
        """Get research location unit ID (pre-DE versions only)."""
        ...

    @location_unit_id.setter
    def location_unit_id(self, value: int) -> None:
        ...

    @property
    def name_str_id(self) -> int:
        """Get language file name string ID."""
        ...

    @name_str_id.setter
    def name_str_id(self, value: int) -> None:
        ...

    @property
    def description_str_id(self) -> int:
        """Get language file description string ID."""
        ...

    @description_str_id.setter
    def description_str_id(self, value: int) -> None:
        ...

    @property
    def help_str_id(self) -> int:
        """Get language file help string ID."""
        ...

    @help_str_id.setter
    def help_str_id(self, value: int) -> None:
        ...

    @property
    def tech_tree_str_id(self) -> int:
        """Get tech tree string ID."""
        ...

    @tech_tree_str_id.setter
    def tech_tree_str_id(self, value: int) -> None:
        ...

    @property
    def hotkey_str_id(self) -> int:
        """Get hotkey string ID (pre-DE versions only)."""
        ...

    @hotkey_str_id.setter
    def hotkey_str_id(self, value: int) -> None:
        ...

    @property
    def button_id(self) -> int:
        """Get button ID (pre-DE versions only)."""
        ...

    @button_id.setter
    def button_id(self, value: int) -> None:
        ...

    @property
    def repeatable(self) -> bool:
        """Get if tech is repeatable (DE only)."""
        ...

    @repeatable.setter
    def repeatable(self, value: bool) -> None:
        ...

    @property
    def name2(self) -> str:
        """Get secondary name (SWGB only)."""
        ...

    @name2.setter
    def name2(self, value: str) -> None:
        ...

    @property
    def research_locations(self) -> list:
        """Get research locations list (DE v8.8+ only)."""
        ...

    def get_research_location(self, location_id: int) -> Optional[Any]:
        """
        Get a specific research location by index.
        
        Args:
            location_id: Index of the research location (0-based)
            
        Returns:
            ResearchLocation object or None if not found
        """
        ...

    def add_research_location(
        self,
        location_unit_id: int = -1,
        research_time: int = 0,
        button_id: int = 0,
        hotkey_str_id: int = -1,
    ) -> Optional[Any]:
        """
        Add a new research location.
        
        Args:
            location_unit_id: Building unit ID where this can be researched
            research_time: Time to research at this location
            button_id: UI button position
            hotkey_str_id: Hotkey string ID
            
        Returns:
            The new ResearchLocation object, or None if failed
        """
        ...

    def remove_research_location(self, location_id: int) -> bool:
        """
        Remove a research location by index.
        
        Args:
            location_id: Index of the research location to remove
            
        Returns:
            True if removed, False if failed or out of range
        """
        ...

    def copy_research_location(self, location_id: int, target_index: Optional[int] = None) -> Optional[Any]:
        """
        Copy a research location.
        
        Args:
            location_id: Source location index
            target_index: Destination index. If None, appends to end.
            
        Returns:
            The new ResearchLocation object, or None if failed
        """
        ...

    def move_research_location(self, source_index: int, target_index: int) -> bool:
        """
        Move a research location to a new position.
        
        Args:
            source_index: Index of location to move
            target_index: New index position
            
        Returns:
            True if moved, False if out of range
        """
        ...

    def clear_research_locations(self) -> None:
        """Remove all research locations."""
        ...

    def exists(self) -> bool:
        """Check if this tech entry exists."""
        ...



from __future__ import annotations

from bfp_rs import BaseStruct, Retriever
from bfp_rs.types.le import u8, f32


class ProjectileInfo(BaseStruct):
    # @formatter:off
    projectile_type: int         = Retriever(u8,  default = 0)
    smart_mode: int              = Retriever(u8,  default = 0)
    hit_mode: int                = Retriever(u8,  default = 0)
    vanish_mode: int             = Retriever(u8,  default = 0)
    area_effect_specials: int    = Retriever(u8,  default = 0)
    projectile_arc: float        = Retriever(f32, default = 0)
    # @formatter:on



"""Graphics module - Sprite/graphic editing functionality."""

from aoe2_genie_tooling.Graphics.graphic_manager import GraphicManager
from aoe2_genie_tooling.Graphics.graphic_handle import GraphicHandle

__all__ = ["GraphicManager", "GraphicHandle"]



"""Type stubs for UnitManager - enables IDE autocomplete"""
from typing import Any, List, Literal, Optional

from aoe2_genie_tooling.Units.unit_handle import UnitHandle


class UnitManager:
    """Manager for creating, cloning, and moving units in a DAT file."""
    
    def create(
        self,
        name: str,
        base_unit_id: Optional[int] = None,
        unit_id: Optional[int] = None,
        enable_for_civs: Optional[List[int]] = None,
        on_conflict: Literal["error", "overwrite"] = "error",
        fill_gaps: Literal["error", "placeholder"] = "placeholder",
    ) -> UnitHandle:
        """
        Create a new unit.
        
        Args:
            name: Name for the new unit
            base_unit_id: Unit ID to clone from. If None, uses first valid unit.
            unit_id: Target unit ID. If None, appends to end.
            enable_for_civs: List of civ IDs to enable for. None = all civs.
            on_conflict: "error" or "overwrite"
            fill_gaps: "error" or "placeholder"
        """
        ...
    
    def clone_into(
        self,
        dest_unit_id: int,
        src_unit_id: int,
        name: Optional[str] = None,
        enable_for_civs: Optional[List[int]] = None,
        on_conflict: Literal["error", "overwrite"] = "error",
        fill_gaps: Literal["error", "placeholder"] = "placeholder",
    ) -> UnitHandle:
        """Clone an existing unit into a specific destination ID."""
        ...
    
    def move(
        self,
        src_unit_id: int,
        dst_unit_id: int,
        on_conflict: Literal["error", "overwrite", "swap"] = "error",
        fill_gaps: Literal["error", "placeholder"] = "placeholder",
    ) -> UnitHandle:
        """Move a unit from source ID to destination ID."""
        ...
    
    def get(self, unit_id: int, civ_ids: Optional[List[int]] = None) -> UnitHandle:
        """Get a handle for an existing unit."""
        ...
    
    def get_unit(self, unit_id: int, civ_id: int = 0) -> Optional[Any]:
        """Get the raw Unit object for a specific unit ID and civ."""
        ...
    
    def exists(self, unit_id: int) -> bool:
        """Check if a unit ID exists (is not None and not a placeholder)."""
        ...
    
    def exists_raw(self, unit_id: int) -> bool:
        """Check if a unit ID exists (is not None, includes placeholders)."""
        ...
    
    def count(self) -> int:
        """Return the total number of unit slots in civ 0."""
        ...
    
    def find_by_name(self, name: str, civ_id: int = 0) -> Optional[UnitHandle]:
        """Find first unit matching name in the specified civ."""
        ...



"""
BuildingWrapper - Building attribute wrapper for UnitHandle.

Ported from building_OLD.py to work with GenieDatParser.

Provides flat property access to BuildingInfo attributes:
- construction_graphic_id, snow_graphic_id, destruction_graphic_id
- adjacent_mode, graphics_angle, disappears_when_built
- stack_unit_id, foundation_terrain_id, completion_tech_id
- garrison_type, garrison_heal_rate, garrison_repair_rate
- sound IDs, annex management, looting_table

Maps to GenieDatParser's BuildingInfo structure.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Optional

if TYPE_CHECKING:
    from sections.civilization.unit import Unit
    from Datasets.attributes import GarrisonType

__all__ = ["BuildingWrapper"]


class BuildingWrapper:
    """
    Wrapper for BuildingInfo attributes.

    Provides flat property access to building-specific stats.
    Changes propagate to all units in the provided list.

    Attributes from GenieDatParser BuildingInfo:
        construction_sprite_id, snow_sprite_id, destruction_sprite_id,
        destruction_rubble_sprite_id, research_sprite_id, research_complete_sprite_id,
        adjacent_mode, graphics_angle, disappears_when_built, stack_unit_id,
        foundation_terrain_id, old_overlay_id, completion_tech_id, can_burn,
        building_annex, head_unit_id, transform_unit_id, transform_sound_id,
        construction_sound_id, wwise_construction_sound_id, wwise_transform_sound_id,
        garrison_type, garrison_heal_rate, garrison_repair_rate,
        salvage_unit_id, salvage_attributes
    """

    __slots__ = ("_units",)

    def __init__(self, units: List["Unit"]) -> None:
        """
        Initialize with list of units to modify.

        Args:
            units: List of Unit objects to proxy
        """
        object.__setattr__(self, "_units", units)

    def _get_building_info(self) -> Optional[Any]:
        """Get BuildingInfo from first unit."""
        if self._units and hasattr(self._units[0], "building_info") and self._units[0].building_info:
            return self._units[0].building_info
        return None

    def _set_all(self, attr: str, value: Any) -> None:
        """Set attribute on all units' building_info."""
        for unit in self._units:
            if hasattr(unit, "building_info") and unit.building_info:
                setattr(unit.building_info, attr, value)

    # -------------------------
    # Graphics
    # -------------------------

    @property
    def construction_graphic_id(self) -> int:
        """Construction graphic ID."""
        bi = self._get_building_info()
        return bi.construction_sprite_id if bi else -1

    @construction_graphic_id.setter
    def construction_graphic_id(self, value: int) -> None:
        self._set_all("construction_sprite_id", value)

    @property
    def snow_graphic_id(self) -> int:
        """Snow variant graphic ID."""
        bi = self._get_building_info()
        return bi.snow_sprite_id if bi else -1

    @snow_graphic_id.setter
    def snow_graphic_id(self, value: int) -> None:
        self._set_all("snow_sprite_id", value)

    @property
    def destruction_graphic_id(self) -> int:
        """Destruction graphic ID."""
        bi = self._get_building_info()
        return bi.destruction_sprite_id if bi else -1

    @destruction_graphic_id.setter
    def destruction_graphic_id(self, value: int) -> None:
        self._set_all("destruction_sprite_id", value)

    @property
    def destruction_rubble_graphic_id(self) -> int:
        """Destruction rubble graphic ID."""
        bi = self._get_building_info()
        return bi.destruction_rubble_sprite_id if bi else -1

    @destruction_rubble_graphic_id.setter
    def destruction_rubble_graphic_id(self, value: int) -> None:
        self._set_all("destruction_rubble_sprite_id", value)

    @property
    def research_graphic_id(self) -> int:
        """Research graphic ID."""
        bi = self._get_building_info()
        return bi.research_sprite_id if bi else -1

    @research_graphic_id.setter
    def research_graphic_id(self, value: int) -> None:
        self._set_all("research_sprite_id", value)

    @property
    def research_complete_graphic_id(self) -> int:
        """Research complete graphic ID."""
        bi = self._get_building_info()
        return bi.research_complete_sprite_id if bi else -1

    @research_complete_graphic_id.setter
    def research_complete_graphic_id(self, value: int) -> None:
        self._set_all("research_complete_sprite_id", value)

    # -------------------------
    # Building Properties
    # -------------------------

    @property
    def adjacent_mode(self) -> int:
        """Adjacent building mode."""
        bi = self._get_building_info()
        return bi.adjacent_mode if bi else 0

    @adjacent_mode.setter
    def adjacent_mode(self, value: int) -> None:
        self._set_all("adjacent_mode", value)

    @property
    def graphics_angle(self) -> int:
        """Graphics angle."""
        bi = self._get_building_info()
        return bi.graphics_angle if bi else 0

    @graphics_angle.setter
    def graphics_angle(self, value: int) -> None:
        self._set_all("graphics_angle", value)

    @property
    def disappears_when_built(self) -> int:
        """Whether building disappears when fully built."""
        bi = self._get_building_info()
        return bi.disappears_when_built if bi else 0

    @disappears_when_built.setter
    def disappears_when_built(self, value: int) -> None:
        self._set_all("disappears_when_built", value)

    @property
    def stack_unit_id(self) -> int:
        """Stack unit ID."""
        bi = self._get_building_info()
        return bi.stack_unit_id if bi else -1

    @stack_unit_id.setter
    def stack_unit_id(self, value: int) -> None:
        self._set_all("stack_unit_id", value)

    @property
    def foundation_terrain_id(self) -> int:
        """Foundation terrain ID."""
        bi = self._get_building_info()
        return bi.foundation_terrain_id if bi else -1

    @foundation_terrain_id.setter
    def foundation_terrain_id(self, value: int) -> None:
        self._set_all("foundation_terrain_id", value)

    @property
    def old_overlap_id(self) -> int:
        """Old overlap ID (old_overlay_id)."""
        bi = self._get_building_info()
        return bi.old_overlay_id if bi else -1

    @old_overlap_id.setter
    def old_overlap_id(self, value: int) -> None:
        self._set_all("old_overlay_id", value)

    @property
    def tech_id(self) -> int:
        """Associated tech ID (completion_tech_id)."""
        bi = self._get_building_info()
        return bi.completion_tech_id if bi else -1

    @tech_id.setter
    def tech_id(self, value: int) -> None:
        self._set_all("completion_tech_id", value)

    @property
    def completion_tech_id(self) -> int:
        """Completion tech ID."""
        bi = self._get_building_info()
        return bi.completion_tech_id if bi else -1

    @completion_tech_id.setter
    def completion_tech_id(self, value: int) -> None:
        self._set_all("completion_tech_id", value)

    @property
    def can_burn(self) -> int:
        """Whether building can burn."""
        bi = self._get_building_info()
        return bi.can_burn if bi else 0

    @can_burn.setter
    def can_burn(self, value: int) -> None:
        self._set_all("can_burn", value)

    # -------------------------
    # Unit References
    # -------------------------

    @property
    def head_unit_id(self) -> int:
        """Head unit ID for annexes."""
        bi = self._get_building_info()
        return bi.head_unit_id if bi else -1

    @head_unit_id.setter
    def head_unit_id(self, value: int) -> None:
        self._set_all("head_unit_id", value)

    @property
    def transform_unit_id(self) -> int:
        """Transform into unit ID."""
        bi = self._get_building_info()
        return bi.transform_unit_id if bi else -1

    @transform_unit_id.setter
    def transform_unit_id(self, value: int) -> None:
        self._set_all("transform_unit_id", value)

    @property
    def salvage_unit_id(self) -> int:
        """Salvage/pile unit ID."""
        bi = self._get_building_info()
        return bi.salvage_unit_id if bi else -1

    @salvage_unit_id.setter
    def salvage_unit_id(self, value: int) -> None:
        self._set_all("salvage_unit_id", value)

    # Backward compatibility alias
    @property
    def pile_unit_id(self) -> int:
        """Pile unit ID (alias for salvage_unit_id)."""
        return self.salvage_unit_id

    @pile_unit_id.setter
    def pile_unit_id(self, value: int) -> None:
        self.salvage_unit_id = value

    # -------------------------
    # Sound Properties
    # -------------------------

    @property
    def transform_sound_id(self) -> int:
        """Transform sound ID."""
        bi = self._get_building_info()
        return bi.transform_sound_id if bi else -1

    @transform_sound_id.setter
    def transform_sound_id(self, value: int) -> None:
        self._set_all("transform_sound_id", value)

    @property
    def construction_sound_id(self) -> int:
        """Construction sound ID."""
        bi = self._get_building_info()
        return bi.construction_sound_id if bi else -1

    @construction_sound_id.setter
    def construction_sound_id(self, value: int) -> None:
        self._set_all("construction_sound_id", value)

    @property
    def wwise_transform_sound_id(self) -> int:
        """Wwise transform sound ID."""
        bi = self._get_building_info()
        return bi.wwise_transform_sound_id if bi else 0

    @wwise_transform_sound_id.setter
    def wwise_transform_sound_id(self, value: int) -> None:
        self._set_all("wwise_transform_sound_id", value)

    @property
    def wwise_construction_sound_id(self) -> int:
        """Wwise construction sound ID."""
        bi = self._get_building_info()
        return bi.wwise_construction_sound_id if bi else 0

    @wwise_construction_sound_id.setter
    def wwise_construction_sound_id(self, value: int) -> None:
        self._set_all("wwise_construction_sound_id", value)

    # -------------------------
    # Garrison Properties
    # -------------------------

    @property
    def garrison_type(self) -> int:
        """Garrison type."""
        bi = self._get_building_info()
        return bi.garrison_type if bi else 0

    @garrison_type.setter
    def garrison_type(self, value: int) -> None:
        self._set_all("garrison_type", value)

    @property
    def garrison_heal_rate(self) -> float:
        """Garrison heal rate."""
        bi = self._get_building_info()
        return bi.garrison_heal_rate if bi else 0.0

    @garrison_heal_rate.setter
    def garrison_heal_rate(self, value: float) -> None:
        self._set_all("garrison_heal_rate", value)

    @property
    def garrison_repair_rate(self) -> float:
        """Garrison repair rate."""
        bi = self._get_building_info()
        return bi.garrison_repair_rate if bi else 0.0

    @garrison_repair_rate.setter
    def garrison_repair_rate(self, value: float) -> None:
        self._set_all("garrison_repair_rate", value)

    # -------------------------
    # Lists (Read-Only)
    # -------------------------

    @property
    def looting_table(self) -> Any:
        """Looting table (salvage_attributes)."""
        bi = self._get_building_info()
        return bi.salvage_attributes if bi else None

    @looting_table.setter
    def looting_table(self, value: Any) -> None:
        """Set the looting table for all units."""
        self._set_all("salvage_attributes", value)

    @property
    def annexes(self) -> "AnnexesManager":
        """Annexes collection manager."""
        return self.annexes_manager

    @annexes.setter
    def annexes(self, value: List) -> None:
        """Set the entire annexes list for all units."""
        for u in self._units:
            if u.building_info:
                u.building_info.building_annex = value

    @property
    def annexes_manager(self) -> "AnnexesManager":
        """
        Get AnnexesManager for managing building annexes.

        Usage:
            building.annexes_manager.set(0, unit_id=500, x=1.0, y=1.0)
            building.annexes_manager[0].unit_id  # Get first annex unit_id
        """
        from aoe2_genie_tooling.Units.unit_collections import AnnexesManager
        return AnnexesManager(self._units)



"""
TrainLocationsManager - Collection manager for unit training locations.

Manages the `creation_info.train_locations_new` collection across multiple units.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Optional, Iterator

from aoe2_genie_tooling.Units.handles import TrainLocationHandle
from sections.civilization.type_info.creation_info import TrainLocation

if TYPE_CHECKING:
    from sections.civilization.unit import Unit

__all__ = ["TrainLocationsManager"]


class TrainLocationsManager:
    """
    Manager for the training locations collection (creation_info.train_locations_new) 
    of a unit bundle.
    """

    __slots__ = ("_units",)

    def __init__(self, units: List[Unit]) -> None:
        object.__setattr__(self, "_units", units)

    def _get_creation_info(self) -> Optional[Any]:
        if self._units and hasattr(self._units[0], "creation_info"):
            return self._units[0].creation_info
        return None

    def __len__(self) -> int:
        ci = self._get_creation_info()
        return len(ci.train_locations_new) if ci and ci.train_locations_new else 0

    def __getitem__(self, index: int) -> TrainLocationHandle:
        ci = self._get_creation_info()
        if ci and 0 <= index < len(ci.train_locations_new):
            return TrainLocationHandle(ci.train_locations_new[index], index)
        raise IndexError(f"Train location index {index} out of range (0-{len(self)-1})")

    def __iter__(self) -> Iterator[TrainLocationHandle]:
        for i in range(len(self)):
            yield self[i]

    def add(
        self,
        unit_id: int,
        train_time: int = 0,
        button_id: int = 0,
        hot_key_id: int = 0,
        **kwargs
    ) -> TrainLocationHandle:
        """Add a training location to all units.
        
        IMPORTANT: Uses setattr with a new list instead of append() because
        bfp_rs Retriever lists share internal storage across clones.
        """
        # Backward compatibility
        if 'hotkey_id' in kwargs:
             hot_key_id = kwargs.pop('hotkey_id')

        loc_idx = -1
        for u in self._units:
            if hasattr(u, "creation_info") and u.creation_info:
                new_loc = TrainLocation(ver=u.ver)
                new_loc.location_unit_id = unit_id
                new_loc.train_time = train_time
                new_loc.button_id = button_id
                new_loc.hotkey_id = hot_key_id
                
                # CRITICAL: Don't use append()! bfp_rs lists share internal storage.
                current_locs = list(u.creation_info.train_locations_new)
                current_locs.append(new_loc)
                u.creation_info.train_locations_new = current_locs  # setattr triggers bfp_rs copy
                
                if loc_idx == -1:
                    loc_idx = len(u.creation_info.train_locations_new) - 1
        
        return self[loc_idx]

    def remove(self, index: int) -> bool:
        """Remove train location at index from all units.
        
        IMPORTANT: Uses setattr with a new list instead of pop() because
        bfp_rs Retriever lists share internal storage across clones.
        """
        removed = False
        for u in self._units:
            if hasattr(u, "creation_info") and u.creation_info:
                current_locs = list(u.creation_info.train_locations_new)
                if 0 <= index < len(current_locs):
                    current_locs.pop(index)
                    u.creation_info.train_locations_new = current_locs  # setattr triggers bfp_rs copy
                    removed = True
        return removed

    def clear(self) -> None:
        """Clear all train locations from all units.
        
        IMPORTANT: Uses setattr with empty list instead of clear() because
        bfp_rs Retriever lists share internal storage across clones.
        """
        for u in self._units:
            if hasattr(u, "creation_info") and u.creation_info:
                u.creation_info.train_locations_new = []  # setattr triggers bfp_rs copy

    def get_by_unit_id(self, unit_id: int) -> Optional[TrainLocationHandle]:
        """Find the first train location for a specific building unit ID."""
        for handle in self:
            if handle.unit_id == unit_id:
                return handle
        return None



"""
Type stubs for TasksManager - Provides IDE autocomplete.
"""
from typing import Iterator, Optional, List, Any
from aoe2_genie_tooling.Units.handles import TaskHandle


class TasksManager:
    """Manager for unit tasks across all civilizations."""
    
    def __init__(self, units: List[Any]) -> None: ...
    
    def __len__(self) -> int:
        """Number of tasks in the primary unit."""
        ...
    
    def __getitem__(self, index: int) -> TaskHandle:
        """Get TaskHandle by index (wrapping tasks from all civs at that index)."""
        ...
    
    def __iter__(self) -> Iterator[TaskHandle]:
        """Iterate over TaskHandles for all tasks."""
        ...
    
    def add(
        self,
        task_type: int = 1,
        id: int = -1,
        is_default: bool = False,
        action_type: int = 0,
        unit_class_id: int = -1,
        unit_type: int = -1,
        terrain_type: int = -1,
        resource_in: int = -1,
        resource_out: int = -1,
        work_value1: float = 0.0,
        work_value2: float = 0.0,
        work_range: float = 0.0,
        **kwargs
    ) -> TaskHandle:
        """Add a new task to all units in the bundle."""
        ...
    
    def remove(self, index: int) -> bool:
        """Remove task at specified index from all units."""
        ...
    
    def clear(self) -> None:
        """Remove all tasks from all units."""
        ...
    
    def get_by_type(self, task_type: int) -> Optional[TaskHandle]:
        """Find the first task with the specified type in the primary unit."""
        ...
    
    def remove_by_action_type(self, action_type: int) -> int:
        """
        Remove all tasks with the specified action_type from all units.
        
        Args:
            action_type: The action type to filter tasks by (e.g., 101 for Build, 133 for Speed Charge).
            
        Returns:
            Number of tasks removed.
        """
        ...



from __future__ import annotations

from bfp_rs import BaseStruct, Retriever
from bfp_rs.types.le import i16


class DamageClass(BaseStruct):
    # @formatter:off
    id: int         = Retriever(i16, default = -1)
    amount: int     = Retriever(i16, default = 0)
    # @formatter:on



"""Type stubs for EffectHandle - enables IDE autocomplete"""
from typing import Any, Optional
from aoe2_genie_tooling.Effects.command_handle import CommandHandle
from aoe2_genie_tooling.Effects.effect_command_builder import EffectCommandBuilder

class EffectHandle:
    """Handle for a single effect holder."""
    
    @property
    def id(self) -> int:
        """Get the effect ID."""
        ...
    
    @property
    def name(self) -> str:
        """Get the effect name."""
        ...

    @name.setter
    def name(self, value: str) -> None:
        ...

    @property
    def commands(self) -> list[CommandHandle]:
        """Get all effect commands inside this holder."""
        ...

    @property
    def effects(self) -> list[CommandHandle]:
        """Alias for commands."""
        ...

    def new_command(
        self,
        type: int = 0,
        a: int = -1,
        b: int = -1,
        c: int = -1,
        d: float = 0.0,
    ) -> CommandHandle:
        """Add a new effect command to this holder (raw method)."""
        ...

    @property
    def add_command(self) -> EffectCommandBuilder:
        """
        Fluent API for adding typed effect commands.
        
        Usage:
            effect.add_command.attribute_modifier_set(a=4, b=-1, c=0, d=100)
            effect.add_command.enable_disable_unit(a=100, b=1)
            effect.add_command.team_upgrade_unit(a=4, b=1000)
        """
        ...

    def get_command(self, index: int) -> Optional[CommandHandle]:
        """Get a command by index."""
        ...

    def copy_command(self, index: int, target_index: Optional[int] = None) -> Optional[CommandHandle]:
        """Copy a command within this holder."""
        ...

    def move_command(self, source_index: int, target_index: int) -> bool:
        """Move a command to a new position within this holder."""
        ...

    def remove_command(self, index: int) -> bool:
        """Remove a command by index."""
        ...

    def clear_commands(self) -> None:
        """Remove all commands."""
        ...

    def exists(self) -> bool:
        """Check if this effect entry exists."""
        ...



"""Techs module - Technology editing functionality."""

from aoe2_genie_tooling.Techs.tech_manager import TechManager
from aoe2_genie_tooling.Techs.tech_handle import TechHandle

__all__ = ["TechManager", "TechHandle"]



from __future__ import annotations

from bfp_rs import BaseStruct, Retriever
from bfp_rs.types.le import i32, i8, i16


class MapLand(BaseStruct):
    # @formatter:off
    id: int                      = Retriever(i32,   default = 0)
    terrain_type: int            = Retriever(i32,   default = 0)
    land_avoidance_distance: int = Retriever(i32,   default = 0)
    base_size: int               = Retriever(i32,   default = 0)
    zone: int                    = Retriever(i8,    default = 0)
    placement_type: int          = Retriever(i8,    default = 0)
    _padding1: int               = Retriever(i16,   default = 0)
    x: int                       = Retriever(i32,   default = 0)
    y: int                       = Retriever(i32,   default = 0)
    land_usage_percent: int      = Retriever(i8,    default = 0)
    by_player_mode: int          = Retriever(i8,    default = 0)
    _padding2: int               = Retriever(i16,   default = 0)
    start_area_radius: int       = Retriever(i32,   default = 0)
    terrain_edge_fade: int       = Retriever(i32,   default = 0)
    clumpiness_factor: int       = Retriever(i32,   default = 0)
    # @formatter:on



from __future__ import annotations

from bfp_rs import BaseStruct, Retriever
from bfp_rs.types.le import i32


class MapTerrain(BaseStruct):
    # @formatter:off
    percent: int            = Retriever(i32,  default = 0)
    type: int               = Retriever(i32,  default = 0)
    num_clumps: int         = Retriever(i32,  default = 0)
    edge_spacing: int       = Retriever(i32,  default = 0)
    placement_zone: int     = Retriever(i32,  default = 0)
    """aka base_terrain_type"""
    clumpiness_factor: int  = Retriever(i32,  default = 0)
    # @formatter:on



"""
Link Finder - Find connections between units above specified thresholds.

This module scans units and finds links (graphics, sounds, unit references)
that are above specified minimum IDs. Only connections above thresholds are reported.

Usage:
    from goku_tools.reverse_engineer.link_finder import LinkFinder
    
    finder = LinkFinder(
        workspace=ws,
        min_unit_id=2300,
        min_graphic_id=2500,
        min_sound_id=700
    )
    
    links = finder.find_links()
"""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Dict, List, Optional, Set, Tuple
from collections import defaultdict

if TYPE_CHECKING:
    from aoe2_genie_tooling.Base.workspace import GenieWorkspace

__all__ = ["LinkFinder", "UnitLinks", "LinkFinderConfig"]


@dataclass
class LinkFinderConfig:
    """Configuration for link finding thresholds."""
    min_unit_id: int = 0        # Only scan units >= this ID
    min_graphic_id: int = 0     # Only report graphic links >= this ID
    min_sound_id: int = 0       # Only report sound links >= this ID
    min_unit_ref_id: int = 0    # Only report unit ref links >= this ID (defaults to min_unit_id)
    
    def __post_init__(self):
        # If min_unit_ref_id not set, default to min_unit_id
        if self.min_unit_ref_id == 0 and self.min_unit_id > 0:
            self.min_unit_ref_id = self.min_unit_id


@dataclass
class UnitLinks:
    """Links found for a single unit."""
    unit_id: int
    name: str
    
    # Filtered links (only those above thresholds)
    graphics: List[int] = field(default_factory=list)
    sounds: List[int] = field(default_factory=list)
    unit_refs: List[int] = field(default_factory=list)
    
    @property
    def has_links(self) -> bool:
        """Check if this unit has any links above thresholds."""
        return bool(self.graphics or self.sounds or self.unit_refs)
    
    @property
    def total_links(self) -> int:
        """Total number of links found."""
        return len(self.graphics) + len(self.sounds) + len(self.unit_refs)


class LinkFinder:
    """
    Find connections/links between units above specified thresholds.
    
    Only reports links where the target ID is >= the corresponding threshold.
    E.g., if min_graphic_id=2500, a unit with standing_graphic=100 will NOT
    be reported, but a unit with standing_graphic=3000 WILL be reported.
    """
    
    def __init__(
        self,
        workspace: "GenieWorkspace",
        min_unit_id: int = 0,
        min_graphic_id: int = 0,
        min_sound_id: int = 0,
        min_unit_ref_id: Optional[int] = None
    ):
        """
        Initialize the link finder.
        
        Args:
            workspace: GenieWorkspace to scan
            min_unit_id: Only scan units with ID >= this value
            min_graphic_id: Only report graphic links >= this value
            min_sound_id: Only report sound links >= this value
            min_unit_ref_id: Only report unit ref links >= this value (defaults to min_unit_id)
        """
        self.ws = workspace
        self.config = LinkFinderConfig(
            min_unit_id=min_unit_id,
            min_graphic_id=min_graphic_id,
            min_sound_id=min_sound_id,
            min_unit_ref_id=min_unit_ref_id if min_unit_ref_id is not None else min_unit_id
        )
        
        # Load field metadata
        from aoe2_genie_tooling.Base.core.field_metadata import get_flat_fields
        self.unit_fields = get_flat_fields("units")
        
        # Categorize fields
        self.graphic_fields = [f for f, ref in self.unit_fields.items() if ref.target_type == "GraphicHandle"]
        self.sound_fields = [f for f, ref in self.unit_fields.items() if ref.target_type == "SoundHandle"]
        self.unit_ref_fields = [f for f, ref in self.unit_fields.items() if ref.target_type == "UnitHandle"]
    
    def _get_field_value(self, obj, field_name: str):
        """Get value from object, traversing dotted paths."""
        if "." in field_name:
            parts = field_name.split(".")
            current = obj
            for part in parts:
                try:
                    current = getattr(current, part, None)
                except Exception:
                    return None
                if current is None:
                    return None
            return current
        return getattr(obj, field_name, None)
    
    def _extract_scalar_ids(self, unit, field_names: List[str]) -> List[int]:
        """Extract scalar IDs from field list."""
        ids = []
        for field in field_names:
            val = self._get_field_value(unit, field)
            if isinstance(val, int) and val >= 0:
                ids.append(val)
        return ids
    
    def _extract_collection_ids(self, unit) -> Tuple[List[int], List[int], List[int]]:
        """Extract IDs from collection fields (tasks, damage_graphics, etc.)."""
        g_ids = []
        s_ids = []
        u_ids = []
        
        # Damage Graphics
        try:
            if unit.damage_graphics:
                for dg in unit.damage_graphics:
                    if hasattr(dg, 'graphic_id') and dg.graphic_id >= 0:
                        g_ids.append(dg.graphic_id)
        except Exception:
            pass
        
        # Tasks
        try:
            if unit.tasks:
                for t in unit.tasks:
                    # Graphics
                    for attr in ['proceeding_graphic_id', 'working_graphic_id', 'carrying_graphic_id']:
                        if hasattr(t, attr):
                            val = getattr(t, attr, -1)
                            if val >= 0:
                                g_ids.append(val)
                    # Sounds
                    for attr in ['resource_gather_sound_id', 'resource_deposit_sound_id']:
                        if hasattr(t, attr):
                            val = getattr(t, attr, -1)
                            if val >= 0:
                                s_ids.append(val)
        except Exception:
            pass
        
        # Train Locations
        try:
            if unit.train_locations:
                for tl in unit.train_locations:
                    if hasattr(tl, 'unit_id') and tl.unit_id >= 0:
                        u_ids.append(tl.unit_id)
        except Exception:
            pass
        
        # Annexes
        try:
            if unit.annexes:
                for a in unit.annexes:
                    if hasattr(a, 'unit_id') and a.unit_id >= 0:
                        u_ids.append(a.unit_id)
        except Exception:
            pass
        
        # Drop Sites
        try:
            if unit.drop_sites:
                for ds in unit.drop_sites:
                    if hasattr(ds, 'unit_id') and ds.unit_id >= 0:
                        u_ids.append(ds.unit_id)
        except Exception:
            pass
        
        return g_ids, s_ids, u_ids
    
    def _discover_graphic_deltas(self, graphic_id: int, visited: Set[int] = None) -> Set[int]:
        """
        Recursively discover all graphics linked via deltas.
        
        Args:
            graphic_id: Starting graphic ID
            visited: Set of already visited graphic IDs (to prevent cycles)
            
        Returns:
            Set of all graphic IDs reachable via delta links
        """
        if visited is None:
            visited = set()
        
        if graphic_id in visited or graphic_id < 0:
            return set()
        
        visited.add(graphic_id)
        result = {graphic_id}
        
        try:
            graphic = self.ws.graphic_manager.get(graphic_id)
            if graphic and hasattr(graphic, 'deltas'):
                for delta in graphic.deltas:
                    delta_gid = int(delta.graphic_id) if delta.graphic_id is not None else -1
                    if delta_gid >= 0 and delta_gid not in visited:
                        # Recursively discover delta-referenced graphics
                        result.update(self._discover_graphic_deltas(delta_gid, visited))
        except Exception:
            pass
        
        return result
    
    def discover_all_graphics(self) -> Set[int]:
        """
        Discover ALL graphics >= min_graphic_id in the workspace.
        Returns set of all graphic IDs above threshold.
        """
        result = set()
        min_gid = self.config.min_graphic_id
        
        try:
            sprites = self.ws.dat.sprites
            for i, sprite in enumerate(sprites):
                if sprite is not None and i >= min_gid:
                    result.add(i)
        except Exception:
            pass
        
        return result
    
    def discover_all_sounds(self) -> Set[int]:
        """
        Discover ALL sounds >= min_sound_id in the workspace.
        Returns set of all sound IDs above threshold.
        """
        result = set()
        min_sid = self.config.min_sound_id
        
        try:
            sounds = self.ws.dat.sounds
            for i, sound in enumerate(sounds):
                if sound is not None and i >= min_sid:
                    result.add(i)
        except Exception:
            pass
        
        return result
    
    def find_links_for_unit(self, unit_id: int) -> Optional[UnitLinks]:
        """
        Find all links for a single unit that are above thresholds.
        
        Returns None if unit doesn't exist or has no links above thresholds.
        """
        try:
            unit = self.ws.unit_manager.get(unit_id)
        except Exception:
            return None
            
        if unit is None or unit._primary_unit is None:
            return None
        
        name = getattr(unit, "name", f"Unit_{unit_id}")
        
        # Extract all IDs
        graphics = self._extract_scalar_ids(unit, self.graphic_fields)
        sounds = self._extract_scalar_ids(unit, self.sound_fields)
        unit_refs = self._extract_scalar_ids(unit, self.unit_ref_fields)
        
        # Add collection IDs
        col_g, col_s, col_u = self._extract_collection_ids(unit)
        graphics.extend(col_g)
        sounds.extend(col_s)
        unit_refs.extend(col_u)
        
        # Filter by thresholds - THIS IS THE KEY LOGIC
        filtered_graphics = [g for g in set(graphics) if g >= self.config.min_graphic_id]
        filtered_sounds = [s for s in set(sounds) if s >= self.config.min_sound_id]
        filtered_unit_refs = [u for u in set(unit_refs) if u >= self.config.min_unit_ref_id]
        
        links = UnitLinks(
            unit_id=unit_id,
            name=name,
            graphics=sorted(filtered_graphics),
            sounds=sorted(filtered_sounds),
            unit_refs=sorted(filtered_unit_refs)
        )
        
        return links if links.has_links else None
    
    def find_links(self) -> Dict[int, UnitLinks]:
        """
        Find all links for units above min_unit_id threshold.
        
        Returns dict of unit_id -> UnitLinks for units that have links above thresholds.
        """
        results = {}
        
        civ = self.ws.dat.civilizations[0]
        total_units = len(civ.units)
        
        for unit_id in range(self.config.min_unit_id, total_units):
            links = self.find_links_for_unit(unit_id)
            if links is not None:
                results[unit_id] = links
        
        return results
    
    def print_summary(self, results: Optional[Dict[int, UnitLinks]] = None):
        """Print a summary of found links."""
        if results is None:
            results = self.find_links()
        
        print(f"\n=== Link Finder Results ===")
        print(f"Config:")
        print(f"  min_unit_id:     {self.config.min_unit_id}")
        print(f"  min_graphic_id:  {self.config.min_graphic_id}")
        print(f"  min_sound_id:    {self.config.min_sound_id}")
        print(f"  min_unit_ref_id: {self.config.min_unit_ref_id}")
        print()
        
        print(f"Units with links above thresholds: {len(results)}")
        print()
        
        # Group by link type
        with_graphics = [u for u in results.values() if u.graphics]
        with_sounds = [u for u in results.values() if u.sounds]
        with_unit_refs = [u for u in results.values() if u.unit_refs]
        
        print(f"  With graphics links: {len(with_graphics)}")
        print(f"  With sound links:    {len(with_sounds)}")
        print(f"  With unit refs:      {len(with_unit_refs)}")
        print()
        
        # Show first 10 units with their links
        print("=== Sample Links (first 10) ===")
        for unit_id, links in list(results.items())[:10]:
            print(f"\n  {unit_id}: {links.name}")
            if links.graphics:
                print(f"    Graphics: {links.graphics}")
            if links.sounds:
                print(f"    Sounds: {links.sounds}")
            if links.unit_refs:
                print(f"    Unit Refs: {links.unit_refs}")


# Quick test runner
if __name__ == "__main__":
    import sys
    sys.path.insert(0, ".")
    from aoe2_genie_tooling.Base.workspace import GenieWorkspace
    
    # Example usage
    print("Loading workspace...")
    ws = GenieWorkspace.load("Testing/GoKu_RPG.dat")
    
    finder = LinkFinder(
        workspace=ws,
        min_unit_id=2300,
        min_graphic_id=15000,
        min_sound_id=700
    )
    
    results = finder.find_links()
    finder.print_summary(results)



"""
AnnexHandle - Wrapper for building annex entries.

Annexes define attached building units (like the corners of a Wonder).
Each annex has a unit_id and positioning info.
"""
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, List, Optional, Tuple

if TYPE_CHECKING:
    # TODO: Removed genie-rust dependency - needs migration to GenieDatParser
    # from genie_rust import Unit
    pass

__all__ = ["AnnexHandle", "AnnexesWrapper"]


@dataclass
class AnnexHandle:
    """
    Handle for a single building annex.

    An annex defines an attached unit that spawns with the building.
    For example, Wonder corner pieces.

    Attributes:
        unit_id: The unit ID to spawn as annex (-1 = none)
        misplacement_x: X offset from building center
        misplacement_y: Y offset from building center
    """
    unit_id: int = -1
    misplacement_x: float = 0.0
    misplacement_y: float = 0.0

    @classmethod
    def from_raw(cls, raw_annex: Any) -> "AnnexHandle":
        """Create from raw genieutils annex object."""
        if raw_annex is None:
            return cls()
        return cls(
            unit_id=getattr(raw_annex, "unit_id", -1),
            misplacement_x=getattr(raw_annex, "misplacement_x", 0.0),
            misplacement_y=getattr(raw_annex, "misplacement_y", 0.0),
        )

    def to_tuple(self) -> Tuple[int, float, float]:
        """Convert to tuple format for storage."""
        return (self.unit_id, self.misplacement_x, self.misplacement_y)

    @property
    def exists(self) -> bool:
        """True if this annex has a valid unit assigned."""
        return self.unit_id >= 0


class AnnexesWrapper:
    """
    Wrapper for managing building annexes.

    Buildings can have up to 4 annexes (attached units).
    This wrapper provides easy access to read/write annexes.

    Usage:
        annexes = unit.building.annexes_wrapper
        annexes[0].unit_id = 500  # Set first annex
        annexes.add(unit_id=501, x=1.0, y=1.0)  # Add annex
    """

    MAX_ANNEXES = 4

    def __init__(self, units: List["Unit"]) -> None:
        """
        Initialize with list of units to modify.

        Args:
            units: List of Unit objects with building component
        """
        self._units = units

    def _get_raw_annexes(self) -> Optional[Tuple]:
        """Get raw annexes tuple from first unit."""
        if self._units and self._units[0].building:
            return self._units[0].building.annexes
        return None

    def __len__(self) -> int:
        """Number of active annexes (with unit_id >= 0)."""
        raw = self._get_raw_annexes()
        if not raw:
            return 0
        return sum(1 for a in raw if hasattr(a, 'unit_id') and a.unit_id >= 0)

    def __getitem__(self, index: int) -> AnnexHandle:
        """Get annex at index (0-3)."""
        if index < 0 or index >= self.MAX_ANNEXES:
            raise IndexError(f"Annex index must be 0-{self.MAX_ANNEXES-1}")

        raw = self._get_raw_annexes()
        if raw and index < len(raw):
            return AnnexHandle.from_raw(raw[index])
        return AnnexHandle()  # Empty annex

    def __iter__(self):
        """Iterate over all annexes."""
        for i in range(self.MAX_ANNEXES):
            yield self[i]

    def all(self) -> List[AnnexHandle]:
        """Get all annexes as list of handles."""
        return [self[i] for i in range(self.MAX_ANNEXES)]

    def active(self) -> List[AnnexHandle]:
        """Get only annexes with valid unit_id."""
        return [a for a in self if a.exists]

    def set(self, index: int, unit_id: int, x: float = 0.0, y: float = 0.0) -> None:
        """
        Set an annex at the given index.

        Args:
            index: Annex slot (0-3)
            unit_id: Unit ID to spawn
            x: X offset from building center
            y: Y offset from building center
        """
        if index < 0 or index >= self.MAX_ANNEXES:
            raise IndexError(f"Annex index must be 0-{self.MAX_ANNEXES-1}")

        for unit in self._units:
            if unit.building and unit.building.annexes:
                annexes = list(unit.building.annexes)
                if index < len(annexes):
                    # Update existing annex
                    annexes[index].unit_id = unit_id
                    annexes[index].misplacement_x = x
                    annexes[index].misplacement_y = y

    def clear(self, index: int) -> None:
        """Clear an annex by setting unit_id to -1."""
        self.set(index, -1, 0.0, 0.0)

    def clear_all(self) -> None:
        """Clear all annexes."""
        for i in range(self.MAX_ANNEXES):
            self.clear(i)



"""
Audit Manifest - Validate and generate review table from manifest.csv.

Usage:
    python audit_manifest.py

Generates manifest_review.md with a formatted table for user approval.
"""
import csv
from pathlib import Path
from typing import Dict, List, Set

# Path constants
DATASETS_DIR = Path(__file__).parent
MANIFEST_PATH = DATASETS_DIR / "manifest.csv"
REVIEW_PATH = DATASETS_DIR / "manifest_review.md"


def load_attribute_enum() -> Dict[int, str]:
    """Load Attribute enum from attributes.py."""
    from attributes import Attribute
    return {attr.value: attr.name for attr in Attribute}


def load_manifest() -> List[Dict[str, str]]:
    """Load manifest.csv and return list of rows."""
    with open(MANIFEST_PATH, newline='', encoding='utf-8') as f:
        reader = csv.DictReader(f)
        return list(reader)


def validate_manifest(manifest: List[Dict], enum_attrs: Dict[int, str]) -> List[str]:
    """Validate manifest against Attribute enum. Returns list of issues."""
    issues = []
    manifest_ids = set()
    
    for row in manifest:
        attr_id = int(row['ID'])
        attr_name = row['Name']
        manifest_ids.add(attr_id)
        
        # Check ID exists in enum
        if attr_id not in enum_attrs:
            issues.append(f"ID {attr_id} ({attr_name}) not in Attribute enum")
        elif enum_attrs[attr_id] != attr_name:
            issues.append(f"ID {attr_id}: name mismatch - manifest='{attr_name}', enum='{enum_attrs[attr_id]}'")
        
        # Check Storage_Type is valid
        if row['Storage_Type'] not in ('Value', 'Reference', 'Bitmask', 'Enum'):
            issues.append(f"ID {attr_id}: invalid Storage_Type '{row['Storage_Type']}'")
        
        # Check Link_Target for References
        if row['Storage_Type'] == 'Reference' and row['Link_Target'] == 'None':
            issues.append(f"ID {attr_id}: Reference type but Link_Target is None")
    
    # Check for missing enum IDs
    for eid, ename in enum_attrs.items():
        if eid not in manifest_ids:
            issues.append(f"Enum ID {eid} ({ename}) missing from manifest")
    
    return issues


def generate_markdown_table(manifest: List[Dict]) -> str:
    """Generate Markdown table from manifest."""
    lines = []
    
    # Header
    lines.append("# Attribute Manifest Review")
    lines.append("")
    lines.append(f"**Total Attributes:** {len(manifest)}")
    lines.append("")
    
    # Summary by Storage_Type
    type_counts = {}
    for row in manifest:
        st = row['Storage_Type']
        type_counts[st] = type_counts.get(st, 0) + 1
    
    lines.append("## Summary by Storage Type")
    lines.append("")
    lines.append("| Storage Type | Count |")
    lines.append("|--------------|-------|")
    for st, count in sorted(type_counts.items()):
        lines.append(f"| {st} | {count} |")
    lines.append("")
    
    # Summary by Link_Target (for References)
    ref_targets = {}
    for row in manifest:
        if row['Storage_Type'] in ('Reference', 'Bitmask', 'Enum'):
            lt = row['Link_Target']
            ref_targets[lt] = ref_targets.get(lt, 0) + 1
    
    lines.append("## Summary by Link Target")
    lines.append("")
    lines.append("| Link Target | Count |")
    lines.append("|-------------|-------|")
    for lt, count in sorted(ref_targets.items()):
        lines.append(f"| {lt} | {count} |")
    lines.append("")
    
    # Full table
    lines.append("## Full Attribute Table")
    lines.append("")
    lines.append("| ID | Name | Storage | Link Target | Type | Description |")
    lines.append("|----|------|---------|-------------|------|-------------|")
    
    for row in manifest:
        lines.append(
            f"| {row['ID']} | {row['Name']} | {row['Storage_Type']} | "
            f"{row['Link_Target']} | {row['Data_Type']} | {row['Description']} |"
        )
    
    return "\n".join(lines)


def main():
    print("Loading Attribute enum...")
    enum_attrs = load_attribute_enum()
    print(f"  Found {len(enum_attrs)} attributes in enum")
    
    print("Loading manifest.csv...")
    manifest = load_manifest()
    print(f"  Found {len(manifest)} rows in manifest")
    
    print("Validating manifest...")
    issues = validate_manifest(manifest, enum_attrs)
    
    if issues:
        print(f"\n⚠ Found {len(issues)} issues:")
        for issue in issues:
            print(f"  - {issue}")
    else:
        print("  ✓ No validation issues")
    
    print("\nGenerating markdown table...")
    markdown = generate_markdown_table(manifest)
    
    REVIEW_PATH.write_text(markdown, encoding='utf-8')
    print(f"  ✓ Written to {REVIEW_PATH}")
    
    print("\nDone! Please review manifest_review.md before proceeding.")


if __name__ == "__main__":
    main()



"""
aoe2_genie_tooling - Production-quality tools layer for editing AoE2 DAT files.

This package provides an AoE2ScenarioParser-style API with GUI-like tabs
and intuitive managers for editing Genie Engine DAT files. It wraps
GenieDatParser (Rust-backed) with a user-friendly, Pythonic interface.

Quick Start:
    from aoe2_genie_tooling import GenieWorkspace

    workspace = GenieWorkspace.load("empires2_x2_p1.dat")
    
    # Access managers through workspace
    unit = workspace.unit_manager.get(4)
    tech = workspace.tech_manager.get(22)
    effect = workspace.effect_manager.get(100)
    
    # Use fluent API for effect commands
    effect.add_command.attribute_modifier_multiply(a=4, b=-1, c=9, d=1.15)
    
    workspace.save("output.dat")

Public API:
    - GenieWorkspace: Root entry point for DAT file editing
    - Managers: UnitManager, TechManager, EffectManager, GraphicManager, SoundManager, CivManager
    - Handles: UnitHandle, TechHandle, EffectHandle, GraphicHandle, SoundHandle, CivHandle
    - logger: Colored console output (can be disabled with logger.disable())
    - registry: JSON export for created items
"""

from aoe2_genie_tooling.Base.workspace import GenieWorkspace
from aoe2_genie_tooling.Base.core.logger import Logger
from aoe2_genie_tooling.Base.core.registry import Registry
from aoe2_genie_tooling.Base.core.exceptions import (
    GenieToolsError,
    GapNotAllowedError,
    InvalidIdError,
    TemplateNotFoundError,
    UnitIdConflictError,
    ValidationError,
)

# Managers
from aoe2_genie_tooling.Units.unit_manager import UnitManager
from aoe2_genie_tooling.Techs.tech_manager import TechManager
from aoe2_genie_tooling.Effects.effect_manager import EffectManager
from aoe2_genie_tooling.Graphics.graphic_manager import GraphicManager
from aoe2_genie_tooling.Sounds.sound_manager import SoundManager
from aoe2_genie_tooling.Civilizations.civ_manager import CivManager

# Handles
from aoe2_genie_tooling.Units.unit_handle import UnitHandle
from aoe2_genie_tooling.Units.task_builder import TaskBuilder
from aoe2_genie_tooling.Techs.tech_handle import TechHandle
from aoe2_genie_tooling.Effects.effect_handle import EffectHandle
from aoe2_genie_tooling.Effects.command_handle import CommandHandle
from aoe2_genie_tooling.Effects.effect_command_builder import EffectCommandBuilder
from aoe2_genie_tooling.Graphics.graphic_handle import GraphicHandle
from aoe2_genie_tooling.Sounds.sound_handle import SoundHandle
from aoe2_genie_tooling.Civilizations.civ_handle import CivHandle

# Create singleton instances for backward compatibility
logger = Logger()
registry = Registry()

__all__ = [
    # Core
    "GenieWorkspace",
    # Managers
    "UnitManager",
    "TechManager",
    "EffectManager",
    "GraphicManager",
    "SoundManager",
    "CivManager",
    # Handles
    "UnitHandle",
    "TaskBuilder",
    "TechHandle",
    "EffectHandle",
    "CommandHandle",
    "EffectCommandBuilder",
    "GraphicHandle",
    "SoundHandle",
    "CivHandle",
    # Logging/Registry
    "logger",
    "registry",
    # Exceptions
    "GenieToolsError",
    "GapNotAllowedError",
    "InvalidIdError",
    "TemplateNotFoundError",
    "UnitIdConflictError",
    "ValidationError",
]

__version__ = "1.3.0"




from __future__ import annotations

from bfp_rs import BaseStruct, Retriever
from bfp_rs.types.le import f32


class AnimationInfo(BaseStruct):
    # @formatter:off
    speed: float = Retriever(f32, default = 0)
    # @formatter:on
